# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import os
import subprocess

with open(os.path.join(os.path.dirname(__file__), '..', 'VERSION')) as \
        version_file:
    release_ = version_file.read().strip()

project = 'cwepr'
copyright = '2020- Mirjam Schröder, 2018/19 Pascal Kirchner, 2018– Till Biskup.'
author = 'Mirjam Schröder, Pascal Kirchner, Till Biskup'

# The short X.Y version
version = ".".join(release_.split('.')[0:2])
# The full version, including alpha/beta/rc tags
release = release_


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    # 'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx_multiversion',
    'sphinxcontrib.bibtex'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Autodoc configuration
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'undoc-members': True,
    'private-members': False,
    'show-inheritance': True,
}

# Multiversion configuration
smv_branch_whitelist = r'^master.*$'
smv_tag_whitelist = r'^v\d+\.\d+$'
smv_released_pattern = r'^refs/tags/v\d+\.\d+$'

tag = subprocess.run("git describe --tags `git rev-list --tags "
                     "--max-count=1`", shell=True, capture_output=True)
smv_latest_version = tag.stdout.decode().strip()

# Bibtex configuration
bibtex_bibfiles = ['literature.bib']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = './cwepr-logo.png'

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = './cwepr-favicon.ico'

html_last_updated_fmt = '%Y-%m-%d'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'cweprdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'cwepr.tex', 'cwepr Documentation',
     author, 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'cwepr', 'cwepr Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'cwepr', 'cwepr Documentation',
     author, 'cwepr', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'aspecd': ('https://docs.aspecd.de/', None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
todo_link_only = True



"""
Plotting: Graphical representations of data extracted from datasets.

Graphical representations of cw-EPR data are an indispensable aspect of data
analysis. To facilitate this, a series of different plotters are available.

Plotting relies on `matplotlib <https://matplotlib.org/>`_, and mainly its
object-oriented interface should be used for the actual plotting.

Generally, two types of plotters can be distinguished:

* Plotters for handling single datasets

  Shall be derived from :class:`aspecd.plotting.SinglePlotter`.

* Plotters for handling multiple datasets

  Shall be derived from :class:`aspecd.plotting.MultiPlotter`.

In the first case, the plot is usually handled using the :meth:`plot` method
of the respective :obj:`cwepr.dataset.Dataset` object. Additionally,
those plotters always only operate on the data of a single dataset, and the
plot can easily be attached as a representation to the respective dataset.
Plotters handling single datasets should always inherit from the
:class:`aspecd.plotting.SinglePlotter` class.

In the second case, the plot is handled using the :meth:`plot` method of the
:obj:`aspecd.plotting.Plotter` object, and the datasets are stored as a list
within the plotter. As these plots span several datasets, there is no easy
connection between a single dataset and such a plot in sense of
representations stored in datasets. Plotters handling multiple datasets should
always inherit from the :class:`aspecd.plotting.MultiPlotter` class.

In a certain sense, there is a third type of plotters:

* Plotters consisting of more than one axes

  Shall be derived from :class:`aspecd.plotting.CompositePlotter`.

However, practically mostly these composite plotters will behave like
plotters handling either single or multiple datasets. Generally,
these composite plotters will use other types of plotters to perform the
actual plot tasks. This modular approach allows for great flexibility.


A note on array dimensions and axes
===================================

Something often quite confusing is the apparent inconsistency between the
order of array dimensions and the order of axes. While we are used to assign
axes in the order *x*, *y*, *z*, and assuming *x* to be horizontal,
*y* vertical (and *z* sticking out of the paper plane), arrays are usually
indexed row-first, column-second. That means, however, that if you simply
plot a 2D array in axes, your *first* dimension is along the *y* axis,
the *second* dimension along the *x* axis.

Therefore, as the axes of your datasets will always correspond to the array
dimensions of your data, in case of 2D plots you will need to *either* use
the information contained in the second axis object for your *x* axis label,
and the information from the first axis object for your *y* axis label,
*or* to transpose the data array.

Another aspect to have in mind is the position of the origin. Usually,
in a Cartesian coordinate system, convention is to have the origin (0,
0) in the *lower left* of the axes (for the positive quadrant). However,
for images, convention is to have the corresponding (0, 0) pixel located in
the *upper left* edge of your image. Therefore, those plotting methods
dealing with images will usually *revert* the direction of your *y* axis.
Most probably, eventually you will have to check with real data and ensure
the plotters to plot data and axes in a consistent fashion.


Types of concrete plotters
==========================

The cwepr package comes with a series of concrete plotters included ready
to be used, thanks to inheriting from the underlying ASpecD framework. As
stated above, plotters can generally be divided into two types: plotters
operating on single datasets and plotters combining the data of multiple
datasets into a single figure.

Additionally, plotters can be categorised with regard to creating figures
consisting of a single or multiple axes. The latter are plotters inheriting
from the :class:`aspecd.plotting.CompositePlotter` class. The latter can be
thought of as templates for the other plotters to operate on, *i.e.* they
provide the axes for other plotters to display their results.


Concrete plotters for single datasets
-------------------------------------

* :class:`cwepr.plotting.SinglePlotter1D`

  Basic line plots for single datasets, allowing to plot a series of
  line-type plots, including (semi)log plots

* :class:`cwepr.plotting.SinglePlotter2D`

  Basic 2D plots for single datasets, allowing to plot a series of 2D plots,
  including contour plots and image-type display

* :class:`aspecd.plotting.SinglePlotter2DStacked`

  Stacked plots of 2D data, converting a 2D display into a series of 1D line
  plots stacked on top of each other.

* :class:`cwepr.plotting.SingleCompositePlotter`

  Composite plotter for single datasets, allowing to plot different views of
  one and the same datasets by using existing plotters for single datasets.

* :class:`cwepr.plotting.GoniometerSweepPlotter`

  Composite plotter for single datasets representing goniometer sweeps,
  *i.e.* angular-dependent cw-EPR measurements.


Concrete plotters for multiple datasets
---------------------------------------

* :class:`cwepr.plotting.MultiPlotter1D`

  Basic line plots for multiple datasets, allowing to plot a series of
  line-type plots, including (semi)log plots

* :class:`cwepr.plotting.MultiPlotter1DStacked`

  Stacked line plots for multiple datasets, allowing to plot a series of
  line-type plots, including (semi)log plots

* :class:`cwepr.plotting.PowerSweepAnalysisPlotter`

  Line plot for multiple datasets particularly for power sweep analysis
  (power saturation analysis) with a second *x* axis on top showing the
  microwave power.


A note for developers
=====================

As each kind of spectroscopy comes with own needs for extensions, there is a
class :class:`PlotterExtensions` that can be used as a mixin class for other
plotters to provide additional functionality for all plotters.

Make sure when implementing functionality here that it really works with all
types of plotters, *i.e.* both SinglePlotters and MultiPlotters. This is
particularly relevant if you need to get information from dataset(s),
as a SinglePlotter will have an attribute ``dataset``, while a MultiPlotter
will have an attribute ``datasets``.


Module documentation
====================

"""
import copy

import numpy as np

import aspecd.plotting
import aspecd.processing

from cwepr import utils


class GoniometerSweepPlotter(aspecd.plotting.SingleCompositePlotter):
    """Overview of the results of a goniometer sweep.

    A goniometer sweep, *i.e.* a series of cw-EPR spectra as a function of
    the angle of the sample with respect to the external magnetic field,
    is usually performed over at least 180°, regardless of the step size.
    The reason is simply that the spectra for 0° and 180° should be
    identical due to the underlying physics of magnetic resonance.

    The plotter will create three subpanels:

    * A 2D plot (scaled image plot) as a general overview.

    * A 1D multiplot comparing the signals for 0° and 180° to check for
      consistency during the measurement.

    * A stacked plot showing all angular positions, providing an alternative
      view of the angular-dependent signal changes compared to the 2D plot.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class.

    To get an overview of your goniometer sweep, just invoke the plotter with
    default values:

    .. code-block:: yaml

       - kind: singleplot
         type: GoniometerSweepPlotter
         properties:
           filename: output.pdf

    """

    def __init__(self):
        super().__init__()
        self.description = 'Plot for one goniometric dataset in different ' \
                           'representations.'
        self.grid_dimensions = [2, 2]
        self.subplot_locations = [[0, 0, 1, 1], [1, 0, 1, 1], [0, 1, 2, 1]]
        self.plotter = [aspecd.plotting.SinglePlotter2D(),
                        aspecd.plotting.MultiPlotter1D(),
                        aspecd.plotting.SinglePlotter2DStacked()]
        self.axes_positions = [[0, 0.15, 1, 1], [0, 0, 1, 1],
                               [0.25, 0, 0.9, 1.07]]
        self.zero_deg_slice = None
        self.hundredeighty_deg_slice = None
        self.parameters['show_zero_lines'] = False
        self.__kind__ = 'singleplot'
        self._exclude_from_to_dict.extend(['dataset', 'zero_deg_slice',
                                           'hundredeighty_deg_slice'])

    def _create_plot(self):
        self._configure_traces_plotter()
        self._configure_contour_plotter()
        self._extract_traces()
        self._configure_comparison_plotter()
        super()._create_plot()

    def _configure_contour_plotter(self):
        upper_contour = self.plotter[0]
        upper_contour.type = 'contourf'
        upper_contour.parameters['show_contour_lines'] = True
        upper_contour.properties.from_dict({
            'axes': {
                'yticks': [0, 30, 60, 90, 120, 150, 180]
            }
        })
        self.plotter[0] = upper_contour

    def _extract_traces(self):
        slicing = aspecd.processing.SliceExtraction()
        slicing.parameters['axis'] = axis_no = 1
        zero_value = self._get_angle_closest_to_value(axis_no, 0)
        hundredeighty_value = self._get_angle_closest_to_value(axis_no, 180)
        slicing.parameters['unit'] = 'axis'
        slicing.parameters['position'] = zero_value
        self.zero_deg_slice = copy.deepcopy(self.dataset)
        self.zero_deg_slice.process(slicing)
        self.zero_deg_slice.label = f'{zero_value:.1f}°'
        slicing.parameters['position'] = hundredeighty_value
        self.hundredeighty_deg_slice = copy.deepcopy(self.dataset)
        self.hundredeighty_deg_slice.process(slicing)
        self.hundredeighty_deg_slice.label = f'{hundredeighty_value:.1f}°'

    def _get_angle_closest_to_value(self, axis_no=0, value=None):
        axis = self.dataset.data.axes[axis_no].values
        return axis[min(range(len(axis)), key=lambda i: abs(axis[i] - value))]

    def _configure_comparison_plotter(self):
        comparison_plotter = self.plotter[1]
        comparison_plotter.datasets = [self.zero_deg_slice,
                                       self.hundredeighty_deg_slice]
        comparison_plotter.properties.from_dict({
            'drawings': [
                {'color': 'tab:blue'},
                {'color': 'tab:red'}
            ],
            'axes': {
                'yticks': [],
                'ylabel': r'$EPR\ intensity$'
            }
        })
        comparison_plotter.parameters['show_legend'] = True
        self.plotter[1] = comparison_plotter

    def _configure_traces_plotter(self):
        self.plotter[2].parameters['yticklabelformat'] = '%.1f'
        self.plotter[2].parameters['ytickcount'] = 19


class PowerSweepAnalysisPlotter(aspecd.plotting.MultiPlotter1D):
    r"""
    Plot results of a power saturation analysis with second axis for mw power.

    To determine the microwave power level not saturating the cw-EPR signal,
    usually a "power sweep" (power saturation study) is carried out with
    systematically varying the incident microwave power. The signal
    amplitude of the resulting data is plotted vs. the square root of the
    microwave power, resulting in a power saturation curve. As long as the
    signal is not saturated, the graph shows a linear relationship.

    As the class inherites from :class:`aspecd.plotting.MultiPlotter1D`
    see there for additional details of the parameters that can be set.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for the plot, implicit and explicit

        Most parameters are documented in the base class. Here, only the
        additional parameters or parameters with specific settings are
        documented.

        mw-axis : class:`bool`
            Whether to show an additional microwave axis in units of power.

            The main *x* axis gives the square root of the microwave power,
            but as the microwave power needs to be set in power units
            (typically mW), it is convenient to have this available as well.

            Default: True

        tight_layout: :class:`bool`
            Whether to adjust the plot to fit into the figure area

            For details see :meth:`matplotlib.figure.Figure.tight_layout`.

            Default: True


    Examples
    --------
    The class basically works like a usual MultiPlotter1D. A full power
    saturation analysis may look like this:

    .. code-block:: yaml

        datasets:
          - PowerSweep
        tasks:
          - kind: singleanalysis
            type: AmplitudeVsPower
            apply_to:
              - PowerSweep
            result: power_sweep_analysis
          - kind: singleanalysis
            type: FitOnData
            properties:
              parameters:
                order: 1
                points: 5
                return_type: dataset
            apply_to:
              - power_sweep_analysis
            result: fit
          - kind: multiplot
            type: PowerSweepAnalysisPlotter
            properties:
              properties:
                drawings:
                  - marker: '*'
                  - color: red
                grid:
                  show: true
                  axis: both
                axes:
                  ylabel: '$EPR\\ amplitude$'
              filename: powersweepanalysis.pdf
            apply_to:
              - power_sweep_analysis
              - fit


    This would result in a power saturation curve (EPR signal amplitude as a
    function of the square root of the microwave power, the latter usually
    in mW), and a linear fit covering in this case the first five data points.

    .. versionadded:: 0.2

    """

    def __init__(self):
        super().__init__()
        self.parameters['mw-axis'] = True
        self.parameters['tight_layout'] = True

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['mw-axis']:
            self._set_lower_xlim()
            self._create_power_axis()

    def _set_lower_xlim(self):
        xlim = list(self.axes.get_xlim())
        if xlim[0] < 0:
            xlim[0] = 0
            self.axes.set_xlim(xlim)

    def _create_power_axis(self):
        """
        Add a mw power axis as second axis opposite the sqrt(mw power) axis.

        Note that :func:`numpy.sqrt` returns NaN for negative values.
        Therefore, the lower axis limit is set to be >= 0 in this plot.
        """
        def forward(values):
            return np.power(values, 2)

        def backward(values):
            return np.sqrt(values)

        power_axis = self.ax.secondary_xaxis('top',
                                             functions=(backward, forward))
        power_axis.set_xlabel('$mw\\ power$')
        power_axis.tick_params(labelrotation=90)


class PlotterExtensions:
    """Extensions for plots of cw-EPR data.

    This class is meant as a mixin class for plotters of the cwepr package
    and provides functionality specific for cw-EPR-spectroscopic data.

    Hence it can only be used as mixin in addition to a plotter class.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for the plot, implicit and explicit

        The following keys exist, in addition to those defined by the actual
        plotter:

        g-axis: :class:`bool`
            Whether to show an additional *g* axis opposite of the magnetic
            field axis

            This assumes the magnetic field axis to be the *x* axis and the
            magnetic field values to be in millitesla (mT), as it calls
            :func:`cwepr.utils.convert_mT2g`.


    .. versionadded:: 0.2

    """

    def __init__(self):
        self.parameters['g-axis'] = False

    def _create_g_axis(self, mw_freq=None):
        """
        Add a *g* axis as second axis opposite the magnetic field axis.

        Currently, this function assumes the magnetic field axis to be the
        *x* axis. Additionally, the magnetic field values are assumed to be
        in millitesla (mT), and the microwave frequency to be in gigahertz (
        GHz).

        Parameters
        ----------
        mw_freq : :class:`float`
            microwave frequency (**in GHz**) used to convert from mT to g

        """
        def forward(values):
            return utils.convert_mT2g(values, mw_freq=mw_freq)

        def backward(values):
            return utils.convert_g2mT(values, mw_freq=mw_freq)

        gaxis = self.ax.secondary_xaxis('top', functions=(backward, forward))
        gaxis.set_xlabel(r'$g\ value$')


class SinglePlotter1D(aspecd.plotting.SinglePlotter1D, PlotterExtensions):
    """1D plots of single datasets.

    Convenience class taking care of 1D plots of single datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SinglePlotter1D`
    class for details.

    Furthermore, the class inhertis all functionality from
    :class:`PlotterExtensions`. See there for additional details.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter1D
         properties:
           filename: output.pdf


    In case you would like to have a *g* axis plotted as a second *x* axis on
    top:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter1D
         properties:
           parameters:
             g-axis: true
           filename: output.pdf

    """

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['g-axis'] and self.dataset.data.axes[0].unit == 'mT':
            self._create_g_axis(self.dataset.metadata.bridge.mw_frequency.value)


class SinglePlotter2D(aspecd.plotting.SinglePlotter2D, PlotterExtensions):
    """2D plots of single datasets.

    Convenience class taking care of 2D plots of single datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SinglePlotter2D`
    class for details.

    Furthermore, the class inhertis all functionality from
    :class:`PlotterExtensions`. See there for additional details.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf

    To change the axes (flip *x* and *y* axis):

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           parameters:
             switch_axes: True

    To use another type (here: contour):

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           type: contour

    To set the number of levels of a contour plot to 10:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           type: contour
           parameters:
             levels: 10

    To change the colormap (cmap) used:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           filename: output.pdf
           properties:
             drawing:
               cmap: RdGy

    Make sure to check the documentation of the ASpecD
    :mod:`aspecd.plotting` module for further parameters that can be set.

    In case you would like to have a *g* axis plotted as a second *x* axis on
    top:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2D
         properties:
           parameters:
             g-axis: true
           filename: output.pdf

    """

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['g-axis'] and self.dataset.data.axes[0].unit == 'mT':
            self._create_g_axis(self.dataset.metadata.bridge.mw_frequency.value)


class SinglePlotter2DStacked(aspecd.plotting.SinglePlotter2DStacked,
                             PlotterExtensions):
    """Stacked plots of 2D data.

    A stackplot creates a series of lines stacked on top of each other from
    a 2D dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.SinglePlotter2DStacked`
    class for details.

    Furthermore, the class inhertis all functionality from
    :class:`PlotterExtensions`. See there for additional details.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2DStacked
         properties:
           filename: output.pdf

    If you need to more precisely control the formatting of the y tick
    labels, particularly the number of decimals shown, you can set the
    formatting accordingly:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2DStacked
         properties:
           filename: output.pdf
           parameters:
             yticklabelformat: '%.2f'

    In this particular case, the y tick labels will appear with only two
    decimals. Note that currently, the "old style" formatting specifications
    are used due to their widespread use in other programming languages and
    hence the familiarity of many users with this particular notation.

    Sometimes you want to have horizontal "zero lines" appear for each
    individual trace of the stacked plot. This can be achieved explicitly
    setting the "show_zero_lines" parameter to "True" that is set to "False"
    by default:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2DStacked
         properties:
           filename: output.pdf
           parameters:
             show_zero_lines: True

    In case you would like to have a *g* axis plotted as a second *x* axis on
    top:

    .. code-block:: yaml

       - kind: singleplot
         type: SinglePlotter2DStacked
         properties:
           parameters:
             g-axis: true
           filename: output.pdf

    """

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['g-axis'] and self.dataset.data.axes[0].unit == 'mT':
            self._create_g_axis(self.dataset.metadata.bridge.mw_frequency.value)


class MultiPlotter1D(aspecd.plotting.MultiPlotter1D, PlotterExtensions):
    """1D plots of multiple datasets.

    Convenience class taking care of 1D plots of multiple datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.MultiPlotter1D`
    class for details.

    Furthermore, the class inhertis all functionality from
    :class:`PlotterExtensions`. See there for additional details.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1D
         properties:
           filename: output.pdf

    To change the settings of each individual line (here the colour and label),
    supposing you have three lines, you need to specify the properties in a
    list for each of the drawings:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1D
         properties:
           filename: output.pdf
           properties:
             drawings:
               - color: '#FF0000'
                 label: foo
               - color: '#00FF00'
                 label: bar
               - color: '#0000FF'
                 label: foobar

    .. important::
        If you set colours using the hexadecimal RGB triple prefixed by
        ``#``, you need to explicitly tell YAML that these are strings,
        surrounding the values by quotation marks.

    In case you would like to have a *g* axis plotted as a second *x* axis on
    top:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1D
         properties:
           parameters:
             g-axis: true
           filename: output.pdf

    """

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['g-axis'] \
                and self.datasets[0].data.axes[0].unit == 'mT':
            self._create_g_axis(
                self.datasets[0].metadata.bridge.mw_frequency.value)


class MultiPlotter1DStacked(aspecd.plotting.MultiPlotter1DStacked,
                            PlotterExtensions):
    """Stacked 1D plots of multiple datasets.

    Convenience class taking care of 1D plots of multiple datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.plotting.MultiPlotter1DStacked`
    class for details.

    Furthermore, the class inhertis all functionality from
    :class:`PlotterExtensions`. See there for additional details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the plotter with default values:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf

    To change the settings of each individual line (here the colour and label),
    supposing you have three lines, you need to specify the properties in a
    list for each of the drawings:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf
           properties:
             drawings:
               - color: '#FF0000'
                 label: foo
               - color: '#00FF00'
                 label: bar
               - color: '#0000FF'
                 label: foobar

    .. important::
        If you set colours using the hexadecimal RGB triple prefixed by
        ``#``, you need to explicitly tell YAML that these are strings,
        surrounding the values by quotation marks.

    Sometimes you want to have horizontal "zero lines" appear for each
    individual trace of the stacked plot. This can be achieved explicitly
    setting the "show_zero_lines" parameter to "True" that is set to "False"
    by default:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           filename: output.pdf
           parameters:
             show_zero_lines: True

    In case you would like to have a *g* axis plotted as a second *x* axis on
    top:

    .. code-block:: yaml

       - kind: multiplot
         type: MultiPlotter1DStacked
         properties:
           parameters:
             g-axis: true
           filename: output.pdf

    """

    def _create_plot(self):
        super()._create_plot()
        if self.parameters['g-axis'] \
                and self.datasets[0].data.axes[0].unit == 'mT':
            self._create_g_axis(
                self.datasets[0].metadata.bridge.mw_frequency.value)



"""Report facilities for cw-EPR data.

Being able to automatically create well-formatted reports using pre-defined
templates opens an entirely new dimension in terms of comparing different
datasets and workflows, besides presenting the results of the research.

This module provides functionality to create reports based on templates
provided either by the user or by the package as such.


"Batteries included": Templates contained in the package
========================================================

The "batteries included" approach of Python itself is probably responsible
to a great deal for the success of Python as a language. Similarly,
the cwepr package tries to provide you with a sensible set of tools you need
for your routine data analysis. Reports are no exception to that rule.

Thanks to being based on the ASpecD framework, the cwepr package comes bundled
with a (growing) series of templates allowing you to create reports of
datasets and alike. Thus, getting access to all information stored in a
single dataset is as simple as calling a single reporter, and in context of
recipe-driven data analysis, it is even simpler:

.. code-block:: yaml

    - kind: report
      type: LaTeXReporter
      properties:
        template: dataset.tex
        filename: report.tex
      compile: true

This would create a report of a dataset that is then stored in the file
``report.tex``, using the template ``dataset.tex`` bundled with the cwepr
(and ASpecD) package. As you even set ``compile`` to true, it would even
compile the LaTeX report, including all figures generated during cooking the
recipe and referenced from within the report. Hence, you end up in your current
directory with both, a LaTeX file ``report.tex`` and a PDF file ``report.pdf``.

For more details, including how to customise reports, have a look at the
documentation of the :mod:`aspecd.report` module of the ASpecD framework.


Notes for developers
====================

.. note::
    The dataset can be given either as dataset in the properties of a recipe, or
    via the apply_to parameter. In the first case, the dataset can be accessed
    in the here implemented reporter class via ``self.dataset`` (and as an
    object), in the latter case indirectly via operating on the
    context-object ``self.context['dataset']`` (and as a dict). The dataset
    has to be given explicitly while the dataset-context is applied
    implicitly. Therefore, here is applied the usage of the context that is
    a bit more complicated in operating but more intuitive to write in recipes.

.. note::
    Still in active development and not fail-safe and easy to use.


Module documentation
====================

"""

import collections
import copy
import os

import aspecd.report
import aspecd.dataset

import cwepr.dataset


class ExperimentalDatasetLaTeXReporter(aspecd.report.LaTeXReporter):
    """Report implementation for cwepr module."""

    def __init__(self, template='', filename=''):
        super().__init__(template=template, filename=filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._tasks = collections.OrderedDict()
        self._figure_name = dict()
        self._exclude_from_to_dict.extend(['dataset'])

    def create(self):
        """Perform all methods to generate a report."""
        self._prepare_metadata()
        self._get_tasks()
        self._get_figure_names()
        self._create_context()
        self.context = self._sanitise_context(self.context)
        super().create()

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _get_tasks(self):
        for task in self.dataset.tasks:
            if task['kind'] in ('analysis', 'processing'):
                self._tasks[(getattr(task['task'],
                                     task['kind']).description)] = {
                    'Parameters': getattr(task['task'],
                                          task['kind']).parameters,
                    'Comment': getattr(task['task'], task['kind']).comment
                }

    def _get_tasks_recursively(self, dict_=None):
        """Gets tasks recursively.

        Is able to get all the tasks performed on an single dataset what is
        only possible thanks to *call by reference* of the history records (
        ?). The dataset initially containing all the data has to be given in
        the recipe explicitly.

        .. note::
            Currently not used

        .. note: :
            Currently, the order of the tasks is *not* preserved because of
            the recursive search and the third step appears before the
            second. Therefore rework this with a different method, probably
            an other *context* keyword in a recipe.

        """
        for task in dict_.tasks:
            if task['kind'] == 'annotation':
                continue
            if task['kind'] == 'analysis':
                if isinstance(task['task'].analysis.result,
                              aspecd.dataset.CalculatedDataset):
                    self._get_tasks_recursively(task['task'].analysis.result)

            self._tasks[(getattr(task['task'],
                                 task['kind']).description)] = {
                'Parameters': getattr(task['task'],
                                      task['kind']).parameters,
                'Comment': getattr(task['task'], task['kind']).comment,
            }

    def _create_context(self):
        """Create a dictionary containing all data to write the report."""
        self.context['TASKS'] = self._tasks
        self.context['METADATA'] = self._metadata
        self.context['FIGURENAMES'] = self.includes

    def _sanitise_context(self, dict_=None):
        """Removes corresponding keys to empty values from context."""
        tmp_dict = copy.deepcopy(dict_)
        for key, value in dict_.items():
            if key == 'dataset':
                continue
            if isinstance(value, (collections.OrderedDict, dict)):
                tmp_dict[key] = self._sanitise_context(value)
            elif not value:
                tmp_dict.pop(key)
        dict_ = tmp_dict
        return dict_

    def _get_figure_names(self):
        """Get the names of the figures used for the report."""
        for i in range(len(self.dataset.representations)):
            if self.dataset.representations[i].plot.description \
                    == '2D plot as scaled image.':
                self._figure_name['Figure2D'] = \
                    self.dataset.representations[i].plot.filename
            elif self.dataset.representations[i].plot.description \
                    == '1D line plot.':
                self._figure_name['Figure1D'] = \
                    self.dataset.representations[i].plot.filename
            else:
                pass


class PowerSweepAnalysisReporter(aspecd.report.LaTeXReporter):
    """Create report for power sweep analysis."""

    def __init__(self, template='', filename=''):
        super().__init__(template=template, filename=filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._tasks = dict()
        self._exclude_from_to_dict.extend(['dataset'])

    def create(self):
        """Perform all methods to generate a report.

        .. todo::
            Do not rely on dataset from recipe properties but use the dataset
            from `apply_to` that is imported into the context. Further deal
            with (meta)data in the context thus that those are easily usable
            in a template.

        """
        self._prepare_metadata()
        #self._get_tasks()
        # TODO: Put figurenames in a dict instead of a list?
        #self._get_figure_names()
        self._create_context()
        self.context = self._sanitise_context(self.context)
        super().create()

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _create_context(self):
        """Create a dictionary containing all data to write the report."""
        self.context['TASKS'] = self._tasks
        self.context['METADATA'] = self._metadata
        self.context['FIGURENAMES'] = self.includes

    def _get_tasks(self):
        for task, _ in enumerate(self.context['dataset']['tasks']):
            task = self.context['dataset']['tasks'][task]
            if task['kind'] in ('analysis', 'processing'):
                self._tasks[task['task'][task['kind']]['description']] = {
                    'Parameters': task['task'][task['kind']]['parameters'],
                    'Comment': task['task'][task['kind']]['comment']
                }
        fit = self.context['FITTING']
        fit_coeffs = fit.metadata.calculation.parameters['coefficients']
        further_tasks = self.context['CALCDATA'].tasks
        self._tasks[further_tasks[0]['task'].analysis.description] = {
            'Parameters': str(fit_coeffs),
            'Comment': further_tasks[0]['task'].analysis.comment
        }

    def _sanitise_context(self, dict_=None):
        tmp_dict = copy.deepcopy(dict_)
        for key, value in dict_.items():
            if key == 'dataset':
                continue
            if isinstance(value, (collections.OrderedDict, dict)):
                tmp_dict[key] = self._sanitise_context(value)
            elif not value:
                tmp_dict.pop(key)
        dict_ = tmp_dict
        return dict_


class DokuwikiCaptionsReporter(aspecd.report.Reporter):
    """Write DokuWiki captions.

    This reporter generates captions containing selected metadata of the
    measurement, that can be directly used in an DokuWiki used e.g. as an
    electronic lab notebook. Only the filename of the uploaded figure still
    has to be inserted.
    """

    def __init__(self, template='', filename=''):
        self.filename = filename
        self.language = 'de'
        self.template = template if template else self._get_template()
        super().__init__(template=self.template, filename=self.filename)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        # private properties
        self._metadata = dict()
        self._figure_name = dict()
        self._exclude_from_to_dict.extend(['dataset'])

    def create(self):
        """Perform all methods to create captions."""
        self._prepare_metadata()
        self._create_context()
        super().create()

    def _get_template(self):
        language = self.language
        module_rootpath = os.path.split(os.path.abspath(__file__))[0]
        return os.path.join(module_rootpath, 'templates', language,
                            'DokuwikiCaption.txt.jinja')

    def _prepare_metadata(self):
        self._metadata = self.context['dataset']['metadata']
        self._metadata['parameter'] = collections.OrderedDict()
        self._collect_experimental_parameters()

    def _collect_experimental_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['sample', 'measurement', 'parameter']:
                self._metadata['parameter'][key] = \
                    self._metadata[key]

    def _create_context(self):
        self.context['METADATA'] = self._metadata


class InfofileReporter(DokuwikiCaptionsReporter):
    """Write infofile with metadata of the dataset.

    As infofiles are written manually by humans, they are inherently
    sensitive to wrong content and incorrect data especially in the
    spectrometer's parameters. As these parameters usually get also
    reported in the respective data file, they are also read from there and
    take precedence in the dataset's metadata.

    Therefore, to correct an infofile, it can be recreated after the dataset
    has been loaded and the metadata sanitised, using this very reporter.

    **Please note:** In order to not write wrong parameters to the infofile
    this way, the :class:`InfofileReporter` needs to be called *immediately*
    after the datasets have been imported, and before any other processing
    takes place that may alter the datasets' metadata.

    Examples
    --------
    .. code-block:: yaml

        - kind: report
          type: InfofileReporter
          properties:
            filename: NewInfofile.info

    """

    def __init__(self):
        self.filename = ''
        self.language = 'en'
        self.template = self._get_template()
        super().__init__(template=self.template, filename=self.filename)
        self._exclude_from_to_dict.extend(['dataset'])

    def _get_template(self):
        language = self.language
        module_rootpath = os.path.split(os.path.abspath(__file__))[0]
        return os.path.join(module_rootpath, 'templates', language,
                            'Infofile.info.jinja')

    def _create_context(self):
        super()._create_context()
        self.context['DATASET_ID'] = \
            os.path.split(self.context['dataset']['id'])[-1]



import glob
import os
import unittest
import aspecd.tasks
import aspecd.processing
import cwepr.analysis
import cwepr.dataset
import cwepr.processing
import cwepr.report

TEST_ROOTPATH = os.path.split(os.path.abspath(__file__))[0]
MODULE_ROOTPATH = os.path.split(os.path.split(os.path.abspath(__file__))[0])[0]


class TestExperimentalDatasetLaTeXReporter(unittest.TestCase):
    def setUp(self):
        source = os.path.join(TEST_ROOTPATH,
                              "io/testdata/test-bes3t-1D-fieldsweep")
        factory = cwepr.dataset.DatasetFactory()
        self.dataset = factory.get_dataset(source=source)
        analysator = cwepr.analysis.Amplitude()
        self.dataset.analyse(analysator)
        algebra = aspecd.processing.ScalarAlgebra()
        algebra.parameters['kind'] = '+'
        algebra.parameters['value'] = 10
        algebra.comment = 'Does this show up in the report?'
        self.dataset.process(algebra)
        self.filename = 'test.tex'
        template_ = os.path.join(
            MODULE_ROOTPATH, 'cwepr', 'templates', 'de', 'report.tex.jinja')
        self.reporter = \
            cwepr.report.ExperimentalDatasetLaTeXReporter(template=template_,
                                                          filename=self.filename)
        self.reporter.context['dataset'] = self.dataset.to_dict()
        self.reporter.dataset = self.dataset

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)
        if os.path.exists(self.filename.replace('tex', 'pdf')):
            os.remove(self.filename.replace('tex', 'pdf'))

    def test_get_tasks(self):
        self.reporter._get_tasks_recursively(self.dataset)

    def test_reporter(self):
        self.reporter.create()
        #self.reporter.compile()

    def test_to_dict_does_not_contain_dataset(self):
        dict_ = self.reporter.to_dict()
        self.assertNotIn('dataset', dict_)


class TestPowerSweepAnalysisReport(unittest.TestCase):
    def setUp(self):
        self.recipe_filename = \
            os.path.join(TEST_ROOTPATH, 'io/testdata/power-sweep-analysis.yaml')
        self.filename = 'PowerSweepReport.tex'
        self.filename2 = 'PowerSweepAnalysis.pdf'
        self.chef = aspecd.tasks.ChefDeService()

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)
        for path in glob.glob(self.recipe_filename.replace('.yaml', '-*.yaml')):
            os.remove(path)
        if os.path.exists(self.filename2):
            os.remove(self.filename2)
        if os.path.exists(self.filename.replace('tex', 'pdf')):
            os.remove(self.filename.replace('tex', 'pdf'))

    def test_reporter(self):
        self.chef.serve(recipe_filename=self.recipe_filename)

    def test_to_dict_does_not_contain_dataset(self):
        reporter = cwepr.report.PowerSweepAnalysisReporter()
        dict_ = reporter.to_dict()
        self.assertNotIn('dataset', dict_)


class TestModulationAmplitudeSweepAnalysisReport(unittest.TestCase):
    def setUp(self):
        self.recipe_filename = \
            os.path.join(TEST_ROOTPATH,
                         'io/testdata/modulation-amplitude-analysis.yaml')
        self.tex_filename = 'ModAmpSweepReport.tex'
        self.plot_filename = 'ModAmpSweepAnalysis.pdf'
        self.chef = aspecd.tasks.ChefDeService()

    def tearDown(self):
        if os.path.exists(self.tex_filename):
            os.remove(self.tex_filename)
        for path in glob.glob(self.recipe_filename.replace('.yaml', '-*.yaml')):
            os.remove(path)
        if os.path.exists(self.plot_filename):
            os.remove(self.plot_filename)
        if os.path.exists(self.tex_filename.replace('tex', 'pdf')):
            os.remove(self.tex_filename.replace('tex', 'pdf'))

    def test_reporter(self):
        self.chef.serve(recipe_filename=self.recipe_filename)


class TestDokuwikiCaptionsReporter(unittest.TestCase):
    def setUp(self):
        self.filename = 'Dokuwiki-caption.txt'
        self.template_ = os.path.join(
            MODULE_ROOTPATH, 'cwepr', 'templates', 'en',
            'DokuwikiCaption.txt.jinja')
        self.reporter = cwepr.report.DokuwikiCaptionsReporter()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        source = \
            os.path.join(TEST_ROOTPATH, "io/testdata/test-magnettech")
        factory = cwepr.dataset.DatasetFactory()
        self.dataset = factory.get_dataset(source=source)
        self.reporter.context['dataset'] = self.dataset.to_dict()

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_reporter(self):
        self.reporter.filename = self.filename
        self.reporter.template = self.template_
        self.reporter.create()
        self.assertTrue(os.path.exists(self.filename))

    def test_reporter_without_template_filename(self):
        self.reporter.filename = self.filename
        self.reporter.create()
        self.assertTrue(os.path.exists(self.filename))

    def test_to_dict_does_not_contain_dataset(self):
        dict_ = self.reporter.to_dict()
        self.assertNotIn('dataset', dict_)


class InfofileReporterBES3T(unittest.TestCase):
    def setUp(self):
        self.filename = 'MyInfofile.info'
        self.template_ = os.path.join(
            MODULE_ROOTPATH, 'cwepr', 'templates', 'en', 'Infofile.info.jinja')
        self.reporter = cwepr.report.InfofileReporter()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        source = \
            os.path.join(TEST_ROOTPATH, "io/testdata/test-bes3t-1D-fieldsweep")
        factory = cwepr.dataset.DatasetFactory()
        self.dataset = factory.get_dataset(source=source)
        self.reporter.context['dataset'] = self.dataset.to_dict()

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_reporter(self):
        self.reporter.filename = self.filename
        self.reporter.template = self.template_
        self.reporter.create()
        self.assertTrue(os.path.exists(self.filename))

    def test_to_dict_does_not_contain_dataset(self):
        dict_ = self.reporter.to_dict()
        self.assertNotIn('dataset', dict_)


class InfofileReporterMagnettech(unittest.TestCase):
    def setUp(self):
        self.filename = 'MyInfofile.info'
        self.template_ = os.path.join(
            MODULE_ROOTPATH, 'cwepr', 'templates', 'en', 'Infofile.info.jinja')
        self.reporter = cwepr.report.InfofileReporter()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.source = \
            os.path.join(TEST_ROOTPATH, "io/testdata/test-magnettech")
        self.factory = cwepr.dataset.DatasetFactory()
        self.source2 = \
            os.path.join(TEST_ROOTPATH, "io/testdata/magnettech-goniometer/")

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_reporter(self):
        for source in (self.source, self.source2):
            with self.subTest(source=source):
                self.dataset = self.factory.get_dataset(source=source)
                self.reporter.context['dataset'] = self.dataset.to_dict()
                self.reporter.filename = self.filename
                self.reporter.template = self.template_
                self.reporter.create()
                self.assertTrue(os.path.exists(self.filename))



"""
Subpackage for IO.

Generally, for each file format (or class of formats), the importer resides
in a separate module.

Currently, a number of more or less well-known and widely used EPR file
formats are supported:

===========================  =============  ==========================
Format                       Extensions     Module
===========================  =============  ==========================
Bruker ESP and EMX           par, spc       :mod:`cwepr.io.esp_winepr`
Bruker BES3T                 DSC, DTA       :mod:`cwepr.io.bes3t`
Magnettech (now Bruker) XML  xml            :mod:`cwepr.io.magnettech`
NIEHS PEST                   lmb, exp, dat  :mod:`cwepr.io.niehs`
---------------------------  -------------  --------------------------
Generic text file            txt            :mod:`cwepr.io.txt_file`
===========================  =============  ==========================

In addition to modules for the individual data file formats, there are a
series of more general modules:

* factory

  Factory classes, currently the DatasetImporterFactory

* exporter

  Exporters, currently only an ASCII exporter


"""

# The import statements below should *only* import the respective classes.
from .factory import DatasetImporterFactory
from .magnettech import MagnettechXMLImporter, GoniometerSweepImporter
from .txt_file import TxtImporter
from .bes3t import BES3TImporter
from .esp_winepr import ESPWinEPRImporter
from .niehs import NIEHSDatImporter, NIEHSLmbImporter, NIEHSExpImporter
from .exporter import ASCIIExporter, MetadataExporter



import collections
import os
import unittest
import cwepr.io.exporter
import cwepr.dataset


class TestMetadataExporter(unittest.TestCase):
    def setUp(self):
        self.export = cwepr.io.exporter.MetadataExporter()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.filename = ''

    def tearDown(self):
        if os.path.exists(self.export.filename):
            os.remove(self.export.filename)

    def test_metadata_dict_is_dict(self):
        self.assertIsInstance(self.export.metadata_dict, collections.OrderedDict)

    def test_file_exists_with_default_filename_after_export(self):
        self.export.dataset = self.dataset
        self.dataset.export_to(self.export)
        self.assertTrue(os.path.exists(self.export.filename))

    def test_file_exists_with_custom_filename_after_export(self):
        self.export.dataset = self.dataset
        self.export.filename = 'my_filename.yaml'
        self.dataset.export_to(self.export)
        self.assertTrue(os.path.exists(self.export.filename))

    def test_file_exists_with_custom_filename_wo_extension_after_export(self):
        self.export.dataset = self.dataset
        self.export.filename = 'my_filename'
        self.dataset.export_to(self.export)
        self.assertTrue(os.path.exists('my_filename.yaml'))

    def test_metadata_get_cleaned_from_empty_values(self):
        self.dataset.metadata.bridge.mw_frequency.value = 9.5
        self.dataset.metadata.bridge.mw_frequency.unit = 'GHz'
        self.export.dataset = self.dataset
        self.export.filename = 'my_filename'
        self.dataset.export_to(self.export)
        goal_dict = {
            'bridge': {
                'mw_frequency': {
                    'value': 9.5,
                    'unit': 'GHz',
                }
            }
        }
        self.assertDictEqual(goal_dict, self.export.metadata_dict)



import datetime
import os
import unittest

import cwepr.dataset
import cwepr.io

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestESPWinEPRImporter(unittest.TestCase):
    def setUp(self):
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.sources = [os.path.join(ROOTPATH, path) for path in [
            'testdata/ESP', 'testdata/EMX-winEPR.par', 'testdata/winepr.par']]
        self.source = os.path.join(ROOTPATH, 'testdata/winepr.par')

    def test_imports_esp_data_correctly(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[0])
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.data[0] < 10 ** 12)

    def test_imports_winepr_data_correctly(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[1])
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.data[0] < 10 ** 12)

    def test_gets_parameter(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[0])
        self.dataset.import_from(importer)
        self.assertTrue(len(importer._par_dict.keys()) > 1)

    def test_infofile_gets_imported(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[0])
        self.dataset.import_from(importer)
        self.assertTrue(isinstance(
            self.dataset.metadata.bridge.mw_frequency.value, float))

    def test_map_par_parameters_correctly(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[0])
        self.dataset.import_from(importer)
        self.assertEqual(5.000000e+05,
                         self.dataset.metadata.signal_channel.receiver_gain
                         .value)
        self.assertAlmostEqual(339.498,
                               self.dataset.metadata.magnetic_field.start.value,
                               2)

    def test_map_par_parameters_correctly_second_dataset(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[1])
        self.dataset.import_from(importer)
        self.assertNotIn('RRG', importer._par_dict.keys())
        self.assertAlmostEqual(350.5,
                               self.dataset.metadata.magnetic_field.start.value,
                               2)

    def test_import_with_1D_dataset(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.source)
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.data.axes[0].unit in ('G', 'mT'))
        self.assertFalse(self.dataset.data.axes[1].unit)

    def test_winepr_sets_default_values(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[2])
        self.dataset.import_from(importer)
        self.assertEqual(100, self.dataset.metadata.signal_channel
                         .modulation_frequency.value)
        self.assertEqual('kHz', self.dataset.metadata.signal_channel
                         .modulation_frequency.unit)

    def test_time_gets_imported_correctly_from_par(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[2])
        self.dataset.import_from(importer)
        date_time = datetime.datetime(2021, 10, 15, 10, 37)
        self.assertEqual(date_time, self.dataset.metadata.measurement.start)

    def test_frequency_gets_written(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[2])
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.metadata.bridge.mw_frequency.value)

    def test_operator_is_written_from_infofile(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[1])
        self.dataset.import_from(importer)
        self.assertTrue(self.dataset.metadata.measurement.operator)
        self.assertTrue(self.dataset.metadata.measurement.purpose)
        self.assertTrue(self.dataset.metadata.experiment.type)
        self.assertTrue(self.dataset.metadata.probehead.type)
        self.assertTrue(self.dataset.metadata.temperature_control.temperature
                        .value)

    def test_mod_amp_has_unit(self):
        importer = cwepr.io.esp_winepr.ESPWinEPRImporter(source=self.sources[1])
        self.dataset.import_from(importer)
        self.assertEqual(
            self.dataset.metadata.signal_channel.modulation_amplitude.unit, 'G')



import unittest

import numpy as np

from cwepr import utils


class TestConvertmT2g(unittest.TestCase):

    def test_values_are_positive(self):
        values = np.linspace(340, 350, 100)
        mw_freq = 9.5
        self.assertTrue(all(utils.convert_mT2g(values, mw_freq) > 0))

    def test_values_have_correct_range(self):
        values = np.linspace(340, 350, 100)
        mw_freq = 9.5
        condition = \
            (np.floor(np.log10(utils.convert_mT2g(values, mw_freq))) == 0)
        self.assertTrue(all(condition))


class TestConvertg2mT(unittest.TestCase):

    def test_values_are_positive(self):
        values = np.linspace(1.8, 4, 100)
        mw_freq = 9.5
        self.assertTrue(all(utils.convert_mT2g(values, mw_freq) > 0))

    def test_values_have_correct_range(self):
        values = np.linspace(1.8, 4, 100)
        mw_freq = 9.5
        condition = \
            (np.floor(np.log10(utils.convert_g2mT(values, mw_freq))) == 2)
        self.assertTrue(all(condition))


class TestNotZero(unittest.TestCase):

    def test_not_zero_of_zero_returns_nonzero_value(self):
        self.assertGreater(utils.not_zero(0), 0)

    def test_not_zero_of_zero_returns_np_float_resolution(self):
        self.assertEqual(np.finfo(np.float64).resolution,
                         utils.not_zero(0))

    def test_not_zero_of_positive_value_preserves_sign(self):
        self.assertGreater(utils.not_zero(1e-20), 0)

    def test_not_zero_of_negative_value_preserves_sign(self):
        self.assertLess(utils.not_zero(-1e-20), 0)

    def test_not_zero_of_negative_value_closer_than_limit_returns_limit(self):
        self.assertEqual(-np.finfo(np.float64).resolution,
                         utils.not_zero(-1e-20))



import aspecd.utils


class_names = ['ExperimentalDataset', 'CalculatedDataset']

for class_name in class_names:
    yaml = aspecd.utils.Yaml()
    ds = aspecd.utils.object_from_class_name(".".join(['cwepr.dataset',
                                                       class_name]))
    yaml.dict = ds.to_dict()
    yaml.serialise_numpy_arrays()
    yaml.write_to(".".join([class_name, 'yaml']))



"""General purpose functions and classes used in other modules.

To avoid circular dependencies, this module does *not* depend on any other
modules of the cwepr package, but it can be imported into every other module.


.. versionadded:: 0.2

"""

import numpy as np
import scipy.constants


def convert_g2mT(values, mw_freq=None):  # noqa
    """
    Convert *g* values to millitesla (mT).

    .. important::

        Magnetic field values are assumed to be in **millitesla** (mT) and
        the microwave frequency to be in **Gigahertz** (GHz).


    Parameters
    ----------
    values : :class:`np.asarray` | :class:`float`
        *g* values to be converted into millitesla (mT)

    mw_freq : :class:`float`
        Microwave frequency (in GHz)

    Returns
    -------
    values : :class:`np.asarray` | :class:`float`
        converted values in millitesla (mT)

    """
    planck_constant = scipy.constants.value('Planck constant')
    mu_b = scipy.constants.value('Bohr magneton')

    values = np.asarray([not_zero(value) for value in values])
    return (planck_constant * mw_freq * 1e9) / (mu_b * values * 1e-3)


def convert_mT2g(values, mw_freq=None):  # noqa
    """
    Convert magnetic field values (in millitesla, mT) to *g* values.

    .. important::

        Magnetic field values are assumed to be in **millitesla** (mT) and
        the microwave frequency to be in **Gigahertz** (GHz).


    Parameters
    ----------
    values : :class:`np.asarray` | :class:`float`
        magnetic field values (in mT) to be converted to *g* values

    mw_freq :
        Microwave frequency (in GHz)

    Returns
    -------
    values : :class:`np.asarray` | :class:`float`
        converted values in *g*

    """
    planck_constant = scipy.constants.value('Planck constant')
    mu_b = scipy.constants.value('Bohr magneton')

    values = np.asarray([not_zero(value) for value in values])
    return (planck_constant * mw_freq * 1e9) / (mu_b * values * 1e-3)


def not_zero(value):
    """
    Return a value that is not zero to prevent DivisionByZero errors.

    Dividing by zero results in NaN values and often hinders evaluating
    mathematical models. A solution adopted from the lmfit Python package
    (https://doi.org/10.5281/zenodo.598352) returns a value equivalent to
    the resolution of a numpy float.

    .. note::

        If you use this function excessively within a module, mostly within
        rather complicated mathematical equations, it might be a good idea
        to import this function explicitly, to shorten the code, such as:
        ``from cwepr.utils import not_zero``. As usual, readability is king.

    Parameters
    ----------
    value : :class:`float`
        Value that can become (too close to) zero to trigger NaN values

    Returns
    -------
    value : :class:`float`
        Value guaranteed not to be zero

    """
    return np.copysign(max(abs(value), np.finfo(np.float64).resolution), value)



r"""Importers for the NIEHS PEST file formats.

.. sidebar:: Contents

    .. contents::
        :local:
        :depth: 1


The US National Institute of Environmental Health Sciences (NIEHS) provides
a collection of public epr software tools, abbreviated PEST. With those
software tools comes a collection of file formats used internally,
both binary and text-based.

For a general overview of PEST, see its homepage:

* https://www.niehs.nih.gov/research/resources/software/tox-pharm/tools/

All these file formats have in common that they contain rather few
metadata. Therefore, it is highly recommended that users provide metadata by
other means, at best in machine-readable format. One possibility would be
to use infofiles, as described in the respective :mod:`aspecd.infofile`
module.


Overview of file formats
========================

The PEST homepage provides an overview of the different file formats
available, for details see:

* https://www.niehs.nih.gov/research/resources/software/tox-pharm/tools/pest/

For convenience, the most important information is provided below:

==============  ============================================================
File extension  Description
==============  ============================================================
     .lmb       NIEHS exclusive binary format
     .exp       ASCII text format with various different realisations
     .dat       ASCII HP-PC interchange format
==============  ============================================================

Just to make things easier, the lmb format exists in two flavours,
with extensions ".lmb" and ".sim", the latter for simulated spectra.

Equally, the ASCII experimental spectra can exist in a plethora of
variations, according to the official documentation: with and without
header, with one or two columns, and with the first column being
magnetic field (in Gauss), time (in seconds), or g value. Interestingly
enough, though, the accepted input format is always two columns with
first column interpreted as magnetic field values.

Details of the individual file formats are given below. Eventually,
it is planned to support all three mentioned file formats.


.. _lmb_format:

NIEHS exclusive binary format (.lmb)
====================================

This file format is a binary format exclusively implemented to be used
with the NIEHS PEST suite. While it is somewhat documented, the best
documentation is still the C source code of the importer used to read the
format. Therefore, below a few more details will be given, together with
the C code used to implement the respective importer.

For authoritative details, the reader is referred to the following online
sources:

* https://www.niehs.nih.gov/research/resources/assets/docs/files.txt
* https://www.niehs.nih.gov/research/resources/assets/docs/files.c
* https://www.niehs.nih.gov/research/resources/assets/docs/constant.h
* https://www.niehs.nih.gov/research/resources/assets/docs/filename.h


General file architecture
-------------------------

The sequence of contents of a file in this format is as follows:

==========  =============  ===================================================
Type        Length         Description
==========  =============  ===================================================
identifier  4 byte         either ``ESRS`` or ``ESR2``
parameters  20x 4 byte     parameters, *e.g.*, for constructing field axis
data        *N*\ x 4 byte  actual data, number of points in parameters[2]
comment     60 byte        comment (in case of ``ESR2`` format, see below)
metadata    20x 12 byte    metadata, *e.g.*, microwave frequency
comments    2x 60 byte     additional comments in case of ``ESR2`` format
==========  =============  ===================================================


Parameter values
----------------

The parameters are not all used. The table below is extracted from the
documentation available at the NIEHS website. Note: Due to a slightly
strange way of using the loops in the original C source code (see below
for details), the number of the parameters in the table below is offset by
one. Furthermore, only those parameters with relevance for the import are
described.

=====  ====================================================================
#      Description
=====  ====================================================================
0      sweep width (in Gauss)
1      center field (in Gauss)
2      number of points
3-8    *unused here*
9      scan time in seconds
10-20  *unused here*
=====  ====================================================================


Metadata
--------
The metadata entries are not all used. The table below is extracted from the
documentation available at the NIEHS website. Note: Due to a slightly
strange way of using the loops in the original C source code (see below
for details), the number of the metadata entries in the table below is
offset by one. Furthermore, only those metadata with relevance for the
import are described.

=====  ====================================================================
#      Description
=====  ====================================================================
0-1    *unused here*
2      modulation amplitude
3      modulation frequency
4      time constant
5      receiver gain
6-7    *unused here*
8      microwave power
9      microwave frequency
10     date (unclear which one: start, end, saving of data)
11     time (unclear which one: start, end, saving of data)
12     scan time in seconds
13     sample temperature
14-20  *unused here*
=====  ====================================================================

.. note::
    Parameters 0-9 are typed in *manually*, therefore, never trust
    these values. Furthermore, due to their human origin, sometimes, value and
    unit are inconsistently separated (no separation, space, ...). Parameters
    10-20 are read from the spectrometer control software. Nevertheless,
    at least for the temperature, this does not mean that the field
    necessarily contains sensible values.


C code of NIEHS lmb importer (reference implementation)
--------------------------------------------------------

The C code shown below is abbreviated (only the important parts of the
files are shown), and it is only for reference. For details, see the
original sources available online from the NIEHS PEST website:

* https://www.niehs.nih.gov/research/resources/assets/docs/files.c
* https://www.niehs.nih.gov/research/resources/assets/docs/constant.h
* https://www.niehs.nih.gov/research/resources/assets/docs/filename.h

Furthermore, the copyright of the source code remains with its original
authors, namely Dave Duling (duling@niehs.nih.gov).


.. code-block:: c
    :caption: Abbreviated contents of the ``files.c`` file containing the
        PEST importer for lmb files. The source code of the two header files
        used is given below.

    /*
       files.c
       basic routines for reading and writing binary epr data files

       updated to use the new binary file format with expanded comments
    */

    #include <stdio.h>
    #include <string.h>
    #include <math.h>

    #include "constant.h"
    #include "filename.h"

    /**************************** READ FILE *****************************

     Read data for an import.  Deals with fpar, fstring, fnote0, and data

     return:   1 if error,   0 if no error.

    */
    int read_file( char *fname, double *ddata, float *fpar,
                   char fstring[][STRSIZE], char *fnote0,
                   char *fnote1, char *fnote2 )
    {

      FILE  *fp;
      int   i, ip;
      float fvalue;
      char  gstring[13];

      /* file not able to be read from */
      if ((fp = fopen(fname, "rb")) == NULL)  return(1);

      fread(gstring, 4, 1, fp);
      gstring[4] = '\0';

      if( strcmp(gstring, FILE_1)!=0 && strcmp(gstring, FILE_2 )!=0 ) {
         fclose(fp);
         return(1);			/* Not correct info. so exit */
      }

      for(i = 1; i <= MAX_PAR; ++i) 	/* read data parameters */
        fread(&fpar[i], 4, 1, fp);

      if( (int) fpar[3] > MAX_PTS ) {	/* check # data points	*/
        fclose(fp);
        return(1);
      }

      for(i = 1; i <= (int)fpar[3]; ++i) {	/* read data values     */
         fread(&fvalue, 4, 1, fp);
         ddata[i] = (double) fvalue;
      }

      fread(fnote0, 60, 1, fp);		/* read comment         */

      for(i = 1; i < STR_NUM ; ++i )	/* read strings         */
        fread( fstring[i], 12, 1, fp );

      if( strcmp(gstring, FILE_2)==0 ) {	/* read extra comments  */
        fread( fnote1, 60, 1, fp );
        fread( fnote2, 60, 1, fp );
      }

      fclose(fp);           /* close file           */
      return(0);            /* return success flag  */

    }  /* end of OPEN FILE */


.. code-block:: c
    :caption: Abbreviated contents of the ``constant.h`` file used in
        ``files.c`` shown above.

    /* constant.h

    */

    /* Define the number of datapoints based on the DOS restriction
       of 64KB of memory in one allocation unit.  Otherwise, use
       many more data points !
    */
    #ifdef DOS_LIMITED
      #define MAX_PTS     4096
      #define MAX_PTS2    2048
    #else
      #define MAX_PTS     16384
      #define MAX_PTS2    8192
    #endif

    /* this data not platform dependent */

    #define MAX_PAR     20
    #define STR_NUM     20
    #define STRSIZE     13
    #define COMMENTSIZE 61


.. code-block:: c
    :caption: Abbreviated contents of the ``filename.h`` file used in
        ``files.c`` shown above.

    /* filename.h

     Filename(s) & info constants
    */

    #define  FILE_1         "ESRS"
    #define  FILE_2         "ESR2"


To import files in this format, use the :class:`NIEHSLmbImporter`. Within
a recipe, the :class:`cwepr.io.factory.DatasetImporterFactory` should
automatically select the correct importer for you.


.. _exp_format:

NIEHS ASCII text format (.exp)
==============================

This text file format is actually a series of different formats, just to
make life easier. The most basic format simply contains (usually two)
columns with numerical data, but there are other versions as well with
additional information on top.

Two example files available from the NIEHS PEST website are documented below,
representing at least two different variants of this format (and currently
the only two variants supported by the respective importer class).


.. code-block::
    :caption: Plain text format containing only columns of data. This
        format is obviously trivial to read. The only tricky part for the
        importer is to distinguish it from the more detailed format shown
        below.

    3290.000	-1.5689E+01
    3290.945	1.5251E+01
    3291.890	2.8689E+00

.. code-block::
    :caption: Text format with additional parameters at the top.
        Presumably the first block contains some description of the data,
        *e.g.* related to a simulation. However, the NIEHS PEST website
        does not provide too much details.

    [EPR]
    N1: DMPO_OOH_OH spectrum S/N=5 with J.Bolton parameters
    N2: DMPO_OH:  aN= 15.3, aH= 15.3, 0.61, 0.25
    N3: DMPO_OOH: aN= 14.9, aH= 14.9

    [DATA]
    3295.10	-0.283
    3295.16	1.523
    3295.22	-0.964


To import files in this format, use the :class:`NIEHSExpImporter`. Within
a recipe, the :class:`cwepr.io.factory.DatasetImporterFactory` should
automatically select the correct importer for you.


.. _dat_format:

ASCII HP-PC interchange format (.dat)
=====================================

This text file format is a rather simple file format with four header
lines and only the intensities stored, but actually no metadata at all. An
example is given below:

.. code-block:: text

    ESRFILE
    50
    3359.27
    8192
    -12.234
       -5.48376
       -2.22887
       0.284066
       -0.698693


The first four lines have special meanings and are referred to as
"header", the following lines (up to the end of the file) are the actual
intensity values of the stored EPR spectrum.

=======  ==============================================================
Content  Description
=======  ==============================================================
ESRFILE  Identifier of the file type
50       scan range in Gauss (here: 50 G)
3359.27  center field in Gauss (here: 3359.27 G)
8192     number of data points (here: 8192) - usually a power of two
=======  ==============================================================

As mentioned above, all following lines contain just the intensity
values. Therefore, the importer needs to reconstruct the field axis
from the center field and scan range values provided. No further
metadata (such as microwave frequency) are available, therefore,
the spectra usually cannot be analysed appropriately unless the
missing information is provided by other means.

To import files in this format, use the :class:`NIEHSDatImporter`. Within
a recipe, the :class:`cwepr.io.factory.DatasetImporterFactory` should
automatically select the correct importer for you.


Module documentation
====================

"""
import io
import struct

import numpy as np

import aspecd.io
import aspecd.annotation


class NIEHSDatImporter(aspecd.io.DatasetImporter):
    """
    Importer for NIEHS PEST dat (text) files.

    The text file format dealt with here is a rather simple file format
    with four header lines and only the intensities stored, but actually
    no metadata at all.

    For details of the file format, see the section :ref:`dat_format`.

    .. note::
        Due to the lack of any metadata of this format, the resulting
        dataset is quite limited in its metadata. Only the axes measures
        and units are added (by informed guessing). Therefore, you are
        highly encouraged to provide the lacking metadata by other means.


    .. versionadded:: 0.3

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self._raw_data = None
        self._raw_metadata = {}

    def _import(self):
        self._clean_filenames()
        self._get_raw_data()
        self._import_data()
        self._create_axis()
        self._assign_units()
        self._assign_some_metadata()

    def _clean_filenames(self):
        # Dirty fix: Cut file extension
        if self.source.endswith((".dat", ".DAT")):
            self.source = self.source[:-4]

    def _get_raw_data(self):
        complete_filename = self.source + ".dat"
        self._raw_data = np.loadtxt(complete_filename, skiprows=1)

    def _import_data(self):
        self.dataset.data.data = self._raw_data[3:]

    def _create_axis(self):
        self._raw_metadata['center_field_mT'] = center_field = \
            self._raw_data[1] / 10
        self._raw_metadata['number_points'] = number_points = \
            int(self._raw_data[2])
        self._raw_metadata['sweep_width_mT'] = sweep_width = \
            self._raw_data[0] / 10
        self._raw_metadata['start'] = center_field - sweep_width / 2
        self._raw_metadata['stop'] = center_field + sweep_width / 2
        axis = np.linspace(center_field - sweep_width / 2,
                           center_field + sweep_width / 2,
                           num=number_points)
        assert (axis[0] == center_field - sweep_width / 2)
        self.dataset.data.axes[0].values = axis

    def _assign_units(self):
        self.dataset.data.axes[0].unit = 'mT'

    def _assign_some_metadata(self):
        self.dataset.metadata.magnetic_field.start.value = \
            self._raw_metadata['start']
        self.dataset.metadata.magnetic_field.stop.value = \
            self._raw_metadata['stop']
        self.dataset.metadata.magnetic_field.points = \
            self._raw_metadata['number_points']
        self.dataset.metadata.magnetic_field.sweep_width.value = \
            self._raw_metadata['sweep_width_mT']
        self.dataset.metadata.magnetic_field.sweep_width.unit = \
            self.dataset.metadata.magnetic_field.start.unit = \
            self.dataset.metadata.magnetic_field.stop.unit = 'mT'


class NIEHSLmbImporter(aspecd.io.DatasetImporter):
    """
    Importer for NIEHS PEST lmb (binary) files.

    This file format is a binary format exclusively implemented to be used
    with the NIEHS PEST suite.

    For details of the file format, see the section :ref:`lmb_format`.

    .. todo::
        Assign further metadata. Note that some metadata are entered
        manually, and therefore, parsing values with units with or without
        space between value and unit can be quite tricky.


    .. versionadded:: 0.3

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self._file_contents = None
        self._file_format = None
        self._parameters = {}
        self._metadata = {}
        self._position = 0
        self._comment = []

    def _import(self):
        self._clean_filenames()
        self._get_raw_data()
        self._detect_file_format()
        self._read_and_assign_parameters()
        self._read_data()
        self._read_comments_and_metadata()
        self._create_axis()
        self._assign_comment()
        self._assign_metadata()

    def _clean_filenames(self):
        # Dirty fix: Cut file extension
        if self.source.endswith((".lmb", ".sim")):
            self.source = self.source[:-4]

    def _get_raw_data(self):
        filename = self.source + ".lmb"
        with open(filename, 'rb') as file:
            self._file_contents = file.read()

    def _detect_file_format(self):
        self._file_format = self._file_contents[:4].decode('utf-8')

    def _read_and_assign_parameters(self):
        parameters = []
        self._position = 4
        max_par = 20
        for _ in range(max_par):
            parameters.append(
                struct.unpack(
                    'f',
                    self._file_contents[self._position:
                                        self._position + 4])[0])
            self._position += 4
        # Note: Only assign those parameters necessary in the given context
        self._parameters = {
            "sweep_width": parameters[0] / 10,
            "center_field": parameters[1] / 10,
            "n_points": int(parameters[2]),
            "scan_time": parameters[9],
        }

    def _read_data(self):
        data = []
        for _ in range(int(self._parameters["n_points"])):
            data.append(
                struct.unpack(
                    'f',
                    self._file_contents[self._position:
                                        self._position + 4])[0])
            self._position += 4
        self.dataset.data.data = np.asarray(data)

    # noinspection GrazieInspection
    def _read_comments_and_metadata(self):
        comment_size = 60
        self._comment = [
            self._file_contents[self._position:
                                self._position + comment_size
                                ].decode('utf-8').replace('\x00', '').strip()]
        self._position += comment_size

        str_num = 20
        str_size = 12
        strings = []
        for _ in range(str_num):
            strings.append(self._file_contents[
                           self._position:
                           self._position + str_size].decode(
                'utf-8').replace('\x00', '').strip())
            self._position += str_size

        if self._file_format == 'ESR2':
            for _ in range(2):
                self._comment.append(
                    self._file_contents[self._position:
                                        self._position + comment_size
                                        ].decode('utf-8').replace('\x00',
                                                                  '').strip())
                self._position += comment_size

        # Only those metadata of interest are mapped
        # Note: The first nine parameters are input MANUALLY, never trust em.
        self._metadata = {
            "modulation_amplitude": strings[1],
            "modulation_frequency": strings[2],
            "time_constant": strings[3],
            "receiver_gain": float(strings[4]),
            "mw_power": strings[7],
            "mw_frequency": strings[8],
            "date": strings[9],
            "time": strings[10],
            "n_scans": int(strings[12]),
            "temperature": strings[13],
        }

    def _create_axis(self):
        self._parameters['start'] = \
            self._parameters['center_field'] \
            - self._parameters['sweep_width'] / 2
        self._parameters['stop'] = \
            self._parameters['center_field'] \
            + self._parameters['sweep_width'] / 2
        axis = np.linspace(self._parameters['start'],
                           self._parameters['stop'],
                           num=self._parameters['n_points'])
        self.dataset.data.axes[0].values = axis
        self.dataset.data.axes[0].unit = 'mT'

    def _assign_comment(self):
        comment_annotation = aspecd.annotation.Comment()
        comment_annotation.comment = self._comment
        self.dataset.annotate(comment_annotation)

    def _assign_metadata(self):
        self.dataset.metadata.magnetic_field.start.value = \
            self.dataset.data.axes[0].values[0]
        self.dataset.metadata.magnetic_field.stop.value = \
            self.dataset.data.axes[0].values[-1]
        self.dataset.metadata.magnetic_field.sweep_width.value = \
            self._parameters["sweep_width"]
        self.dataset.metadata.magnetic_field.start.unit = \
            self.dataset.metadata.magnetic_field.stop.unit = \
            self.dataset.metadata.magnetic_field.sweep_width.unit = "mT"
        self.dataset.metadata.signal_channel.accumulations = \
            self._metadata["n_scans"]


class NIEHSExpImporter(aspecd.io.DatasetImporter):
    """
    Importer for NIEHS PEST exp (text) files.

    The text file format dealt with here is actually a series of different
    formats, just to make life easier. The most basic format simply
    contains (usually two) columns with numerical data, but there are
    other versions as well with additional information on top.

    In case of the latter format with additional information on top of the
    file, this information is added as comment annotation to the dataset.

    For details of the file format, see the section :ref:`exp_format`.

    .. note::
        Due to the lack of any metadata of this format, the resulting
        dataset is quite limited in its metadata. Only the axes measures
        and units are added (by informed guessing). Therefore, you are
        highly encouraged to provide the lacking metadata by other means.


    .. versionadded:: 0.3

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self._file_contents = None
        self._lines = None
        self._file_format = None
        self._raw_data = None
        self._header = None

    def _import(self):
        self._clean_filenames()
        self._read_file_contents()
        self._detect_file_format()
        self._import_data()
        self._assign_data_and_axis()
        self._assign_comment()

    def _clean_filenames(self):
        # Dirty fix: Cut file extension
        if self.source.endswith(".exp"):
            self.source = self.source[:-4]

    def _read_file_contents(self):
        filename = self.source + ".exp"
        with open(filename, 'r') as file:
            self._file_contents = file.read()
        self._lines = self._file_contents.splitlines()

    def _detect_file_format(self):
        if self._lines[0].startswith('['):
            self._file_format = 'with_blocks'
        else:
            self._file_format = 'DSV'

    def _import_data(self):
        if self._file_format == 'with_blocks':
            # noinspection PyTypeChecker
            self._raw_data = np.loadtxt(
                io.StringIO(self._file_contents),
                skiprows=self._lines.index('[DATA]') + 1)
            self._header = self._lines[:self._lines.index('[DATA]')]
        else:
            # noinspection PyTypeChecker
            self._raw_data = np.loadtxt(io.StringIO(self._file_contents))

    def _assign_data_and_axis(self):
        self.dataset.data.axes[0].values = self._raw_data[:, 0] / 10
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.data = self._raw_data[:, 1]

    def _assign_comment(self):
        if self._header:
            # Remove empty lines and trailing spaces
            self._header = [element.rstrip() for element in self._header
                            if element != '']
            comment_annotation = aspecd.annotation.Comment()
            comment_annotation.comment = self._header
            self.dataset.annotate(comment_annotation)



import os
import unittest

import matplotlib
import numpy as np

from aspecd.plotting import Saver

import cwepr.plotting
import cwepr.dataset
import cwepr.io.magnettech

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestGoniometerSweepPlotter(unittest.TestCase):
    def setUp(self):
        self.filename = 'goniometertest.pdf'
        source = os.path.join(ROOTPATH, 'io/testdata/magnettech-goniometer/')
        self.importer = cwepr.io.magnettech.GoniometerSweepImporter(
            source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(self.importer)

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_plotter_does_not_fail(self):
        plotter = cwepr.plotting.GoniometerSweepPlotter()
        plotter.dataset = self.dataset
        plotter.properties.axes.xlim = [337.5, 339]
        saver = Saver()
        saver.filename = self.filename
        self.dataset.plot(plotter)
        plotter.save(saver)


class TestSinglePlotter1D(unittest.TestCase):

    def setUp(self):
        self.plotter = cwepr.plotting.SinglePlotter1D()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random(5)
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'
        self.dataset.data.axes[1].unit = 'V'
        self.plotter.dataset = self.dataset

    def test_has_g_axis_parameter(self):
        self.assertTrue('g-axis' in self.plotter.parameters)

    def test_g_axis_adds_secondary_axis(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertTrue(secondary_axes)

    def test_g_axis_has_correct_label(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertIn('g\\ value',
                      secondary_axes[0].get_xaxis().get_label().get_text())


class TestSinglePlotter2D(unittest.TestCase):

    def setUp(self):
        self.plotter = cwepr.plotting.SinglePlotter2D()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random([5, 5])
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'time'
        self.dataset.data.axes[1].unit = 's'
        self.dataset.data.axes[2].quantity = 'intensity'
        self.dataset.data.axes[2].unit = 'V'
        self.plotter.dataset = self.dataset

    def test_has_g_axis_parameter(self):
        self.assertTrue('g-axis' in self.plotter.parameters)

    def test_g_axis_adds_secondary_axis(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertTrue(secondary_axes)

    def test_g_axis_has_correct_label(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertIn('g\\ value',
                      secondary_axes[0].get_xaxis().get_label().get_text())


class TestSinglePlotter2DStacked(unittest.TestCase):

    def setUp(self):
        self.plotter = cwepr.plotting.SinglePlotter2DStacked()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random([5, 5])
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'time'
        self.dataset.data.axes[1].unit = 's'
        self.dataset.data.axes[2].quantity = 'intensity'
        self.dataset.data.axes[2].unit = 'V'
        self.plotter.dataset = self.dataset

    def test_has_g_axis_parameter(self):
        self.assertTrue('g-axis' in self.plotter.parameters)

    def test_g_axis_adds_secondary_axis(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertTrue(secondary_axes)

    def test_g_axis_has_correct_label(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertIn('g\\ value',
                      secondary_axes[0].get_xaxis().get_label().get_text())


class TestMultiPlotter1D(unittest.TestCase):

    def setUp(self):
        self.plotter = cwepr.plotting.MultiPlotter1D()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random(5)
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'
        self.dataset.data.axes[1].unit = 'V'
        self.plotter.datasets = [self.dataset]

    def test_has_g_axis_parameter(self):
        self.assertTrue('g-axis' in self.plotter.parameters)

    def test_g_axis_adds_secondary_axis(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertTrue(secondary_axes)

    def test_g_axis_has_correct_label(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertIn('g\\ value',
                      secondary_axes[0].get_xaxis().get_label().get_text())


class TestMultiPlotter1DStacked(unittest.TestCase):

    def setUp(self):
        self.plotter = cwepr.plotting.MultiPlotter1DStacked()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random(5)
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'
        self.dataset.data.axes[1].unit = 'V'
        self.plotter.datasets = [self.dataset]

    def test_has_g_axis_parameter(self):
        self.assertTrue('g-axis' in self.plotter.parameters)

    def test_g_axis_adds_secondary_axis(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertTrue(secondary_axes)

    def test_g_axis_has_correct_label(self):
        self.plotter.parameters['g-axis'] = True
        self.plotter.plot()
        secondary_axes = [
            child for child in self.plotter.ax.get_children()
            if isinstance(child, matplotlib.axes._secondary_axes.SecondaryAxis)
        ]
        self.assertIn('g\\ value',
                      secondary_axes[0].get_xaxis().get_label().get_text())



"""
Importer for simple text and CSV files.

Despite its obvious limitations (normally no metadata, limited resolution,
large files) text or CSV files are an often-encountered exchange format and
can serve as an easy way to import data from otherwise not supported file
formats, as nearly every software to record data allows to export these data
as simple text or CSV files.

You may have a look as well at the importers provided by the ASpecD package
for similar situations, particularly :class:`aspecd.io.TxtImporter`.

"""
import numpy as np

import aspecd.io


class TxtImporter(aspecd.io.DatasetImporter):
    """Simple importer for txt files containing data.

    Sometimes, data come from sources as a txt file only. So far, the format
    is hardcoded and should contain two columns separated by a tabulator.

    """

    def __init__(self, source=''):
        super().__init__(source=source)
        # public properties
        self.extension = '.txt'

    def _import(self):
        self._get_data()
        self._create_metadata()

    def _get_data(self):
        self.source = self.source + self.extension
        raw_data = np.loadtxt(self.source, delimiter='\t')
        self.dataset.data.data = raw_data[:, 1]
        self.dataset.data.axes[0].values = raw_data[:, 0]

    def _create_metadata(self):
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'


class CsvImporter(aspecd.io.DatasetImporter):
    """Importer for simple csv imports with different delimiters."""

    def __init__(self, source=''):
        super().__init__(source=source)
        # public properties
        self.extension = '.csv'

    def _import(self):
        self._read_data()

    def _read_data(self):
        with open(self.source + self.extension) as csv_file:
            # noinspection PyTypeChecker
            raw_data = np.loadtxt(csv_file, delimiter=';', skiprows=3)
            self.dataset.data.data = raw_data[:, 1]
            self.dataset.data.axes[0].values = raw_data[:, 0]

    def _create_metadata(self):
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[1].quantity = 'intensity'



import copy
import os
import unittest

import aspecd.exceptions
import numpy as np

import cwepr.exceptions
import cwepr.processing
import cwepr.dataset
import cwepr.io.magnettech

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestAutomaticPhaseCorrection(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/phase-45_noise0.000')
        importer = cwepr.io.txt_file.TxtImporter(source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(importer)

    def test_analytic_signal_is_complex(self):
        apc = cwepr.processing.AutomaticPhaseCorrection()
        apc.parameters['order'] = 1
        apc.parameters['points_percentage'] = 5
        self.dataset.process(apc)
        self.assertTrue(np.iscomplex(apc._analytic_signal).all)

    def test_signal_before_and_after_differ(self):
        apc = cwepr.processing.AutomaticPhaseCorrection()
        apc.parameters['order'] = 1
        apc.parameters['points_percentage'] = 20
        dataset_old = copy.deepcopy(self.dataset)
        self.dataset.process(apc)
        self.assertTrue((dataset_old.data.data != self.dataset.data.data).all())


class TestSubtraction(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/test-magnettech')
        importer = cwepr.io.magnettech.MagnettechXMLImporter(source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(importer)


class TestFrequencyCorrection(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/test-magnettech')
        importer = cwepr.io.magnettech.MagnettechXMLImporter(source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(importer)
        self.corrector = cwepr.processing.FrequencyCorrection()

    def test_frequency_before_is_different_from_after(self):
        old_freq = copy.deepcopy(self.dataset.metadata.bridge.mw_frequency)
        self.corrector.parameters['frequency'] = 9.5
        self.dataset.process(self.corrector)
        new_freq = self.dataset.metadata.bridge.mw_frequency
        self.assertNotEqual(new_freq.value, old_freq.value)

    def test_magnetic_field_axis_is_different(self):
        old_field_axis = copy.deepcopy(
            self.dataset.data.axes[0].values)
        self.corrector.parameters['frequency'] = 8.
        self.dataset.process(self.corrector)
        new_field_axis = self.dataset.data.axes[0].values
        diffs = old_field_axis - new_field_axis
        conditions = (diff == 0 for diff in diffs)
        self.assertFalse(all(conditions))


class GAxisCreation(unittest.TestCase):
    def setUp(self):
        self.proc = cwepr.processing.GAxisCreation()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.data.data = np.random.random(100)
        self.dataset.data.axes[0].values = np.linspace(300, 400, num=100)
        self.dataset.data.axes[0].unit = 'mT'
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.metadata.bridge.mw_frequency.value = 9.5

    def test_instantiate_class(self):
        proc = cwepr.processing.GAxisCreation()
        self.assertTrue(proc.description)

    def test_description_is_appropriate(self):
        self.assertTrue(self.proc.description)
        self.assertIn('magnetic field axis to g axis', self.proc.description)

    def test_axis_values_differs_after(self):
        values_before = np.copy(self.dataset.data.axes[0].values)
        self.dataset.process(self.proc)
        values_after = self.dataset.data.axes[0].values
        diffs = values_before - values_after
        conditions = (diff == 0 for diff in diffs)
        self.assertFalse(all(conditions))

    def test_axis_values_are_positive_values(self):
        self.dataset.process(self.proc)
        self.assertTrue(all(self.dataset.data.axes[0].values > 0))

    def test_axis_values_have_correct_range(self):
        self.dataset.process(self.proc)
        condition = np.floor(np.log10(self.dataset.data.axes[0].values)) == 0
        self.assertTrue(all(condition))

    def test_axis_unit_gets_removed(self):
        self.dataset.process(self.proc)
        self.assertEqual("", self.dataset.data.axes[0].unit)

    def test_axis_quantity_gets_set(self):
        self.dataset.process(self.proc)
        self.assertEqual("g value", self.dataset.data.axes[0].quantity)


class TestAxisInterpolation(unittest.TestCase):
    def setUp(self):
        self.interpolator = cwepr.processing.AxisInterpolation()

    def test_instantiate_class(self):
        pass

    def test_interpolate_returns_new_axis(self):
        source = os.path.join(ROOTPATH, 'io/testdata/test-magnettech')
        importer = cwepr.io.magnettech.MagnettechXMLImporter(source=source)
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(importer)
        old_axis = dataset.data.axes[0].values
        corrector = cwepr.processing.AxisInterpolation()
        dataset.process(corrector)
        new_axis = dataset.data.axes[0].values
        self.assertTrue(np.not_equal(new_axis, old_axis).all())


class TestNormalisationOfDerivativeToArea(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/not_noisy_data')
        importer = cwepr.io.txt_file.TxtImporter(source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(importer)

    def test_normalisation_of_derivative_to_area(self):
        area = cwepr.processing.NormalisationOfDerivativeToArea()
        self.dataset.process(area)
        self.assertEqual(float, type(area._area))
        self.assertTrue(max(self.dataset.data.data) < 1)


class TestNormalisation(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/BDPA-1DFieldSweep')
        importer = cwepr.dataset.DatasetFactory()
        self.dataset = importer.get_dataset(source=source)

    def test_normalisation_to_receiver_gain(self):
        correction = cwepr.processing.Normalisation()
        correction.parameters['kind'] = 'receiver_gain'
        before = max(self.dataset.data.data)
        rg = 10**(self.dataset.metadata.signal_channel.receiver_gain.value /20)
        self.dataset.process(correction)
        after = max(self.dataset.data.data)
        self.assertEqual(before/rg, after)

    def test_normalisation_to_scan_number(self):
        correction = cwepr.processing.Normalisation()
        correction.parameters['kind'] = 'scan_number'
        before = max(self.dataset.data.data)
        scans = self.dataset.metadata.signal_channel.accumulations
        self.dataset.process(correction)
        after = max(self.dataset.data.data)
        self.assertEqual(before/scans, after)

    def test_normalisation_to_maximum(self):
        correction = cwepr.processing.Normalisation()
        correction.parameters['kind'] = 'max'
        before = np.max(self.dataset.data.data)
        max_ = max(self.dataset.data.data)
        self.dataset.process(correction)
        after = max(self.dataset.data.data)
        self.assertEqual(before/max_, after)

    def test_normalisation_to_minimum(self):
        correction = cwepr.processing.Normalisation()
        correction.parameters['kind'] = 'min'
        before = min(self.dataset.data.data)
        min_ = min(self.dataset.data.data)
        self.dataset.process(correction)
        after = min(self.dataset.data.data)
        self.assertEqual(before/abs(min_), after)



import shutil
import tempfile
import unittest
import os

import aspecd.io

import cwepr.dataset
import cwepr.exceptions
import cwepr.io

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestDatasetImporterFactory(unittest.TestCase):
    def setUp(self):
        self.factory = cwepr.io.DatasetImporterFactory()

    def test_instantiate_class(self):
        pass

    def test_goniometer_importer_gets_correct_files(self):
        source = os.path.join(ROOTPATH, 'testdata', 'magnettech-goniometer/')
        importer = cwepr.dataset.DatasetFactory().importer_factory
        importer.get_importer(source=source)
        self.assertEqual('GoniometerSweep', importer.data_format)

    def test_goniometer_importer_does_not_import_inconsistent_data(self):
        source = os.path.join(ROOTPATH, 'testdata', 'magnettech-goniometer/')
        with tempfile.TemporaryDirectory() as tmpdir:
            new_source = os.path.join(tmpdir, 'new')
            shutil.copytree(source, new_source)
            list_element = os.listdir(new_source)[2]
            source_name = os.path.join(new_source, list_element)
            target_name = os.path.join(new_source, 'fake_name')
            os.rename(source_name, target_name)
            importer = cwepr.dataset.DatasetFactory().importer_factory
            importer.get_importer(source=new_source)
            self.assertNotEqual('MagnettechXML', importer.data_format)

    def test_factory_cuts_filename(self):
        source = os.path.join(ROOTPATH, 'testdata', 'test-magnettech.xml')
        factory = cwepr.dataset.DatasetFactory()
        importer_factory = factory.importer_factory
        importer_factory.get_importer(source=source)
        root_source, _ = os.path.splitext(source)
        self.assertEqual(root_source, importer_factory.source)

    def test_with_adf_extension_returns_adf_importer(self):
        source = 'test.adf'
        importer = self.factory.get_importer(source=source)
        self.assertIsInstance(importer, aspecd.io.AdfImporter)

    def test_niehsdat_file_returns_correct_importer(self):
        source = os.path.join(ROOTPATH, 'testdata', 'Pyrene.dat')
        importer = self.factory.get_importer(source=source)
        self.assertIsInstance(importer, cwepr.io.NIEHSDatImporter)

    def test_niehslmb_file_returns_correct_importer(self):
        source = os.path.join(ROOTPATH, 'testdata', 'dmpo.lmb')
        importer = self.factory.get_importer(source=source)
        self.assertIsInstance(importer, cwepr.io.NIEHSLmbImporter)

    def test_niehsexp_file_returns_correct_importer(self):
        source = os.path.join(ROOTPATH, 'testdata', 'e1-05.exp')
        importer = self.factory.get_importer(source=source)
        self.assertIsInstance(importer, cwepr.io.NIEHSExpImporter)



"""Exporters for both, whole datasets as well as data alone.

Exporters
=========

There are two different general types of exporters provided: Exporters for
whole datasets i.e. data with their according metadata and exporters for data
only.

    .. warning::
        All metadata contained within a dataset (including the full history)
        are lost when exporting to plain text. Therefore, using this
        exporter will usually result in you loosing reproducibility. Hence,
        better think twice before using this exporter and use entirely on
        your own risk and only if you *really* know what you are doing (and
        why).

For further details and a comprehensive summary of data export, the reader is
referred to the aspecd-documentation (:mod:`aspecd.io`).

If you want to export data to a txt-file, you might want to use the
:class:`aspecd.io.TxtExporter`.
"""
import collections
import datetime

import numpy as np
import aspecd.io
import aspecd.utils


class ASCIIExporter(aspecd.io.DatasetExporter):
    """Export a dataset in ASCII format.

    Exports the complete dataset to an ASCII file. At the same time, the
    respective metadata is exported into a YAML file using the functionality
    provided by aspecd. Metadata residing there as an object of the class
    :class:`numpy.ndarray` are converted into lists.

    A simple example for its usage looks like this:

    .. code-block:: yaml

        datasets:
          - /path/to/first/dataset
          - /path/to/second/dataset

        tasks:
          - kind: export
            type: ASCIIExporter
            properties:
              target:
                - dataset1
                - dataset2


    .. note::
        Not tested.
    """

    def __init__(self):
        super().__init__()

    def _export(self):
        """Export the dataset's numeric data and metadata."""
        file_name_data = self.target + ".txt"
        file_name_meta = self.target + ".yaml"
        np.savetxt(file_name_data, self.dataset.data.data, delimiter=",")
        metadata_writer = aspecd.utils.Yaml()
        metadata = self._get_and_prepare_metadata()
        metadata_writer.dict = metadata
        metadata_writer.write_to(filename=file_name_meta)

    def _get_and_prepare_metadata(self):
        """Prepare the dataset's metadata to be imported.

        Transforms the metadata to a dict and subsequently transforms all
        instances of :class:`numpy.ndarray` by transforming them into lists.

        Returns
        -------
        metadata_prepared: :class:`dict`
            transformed metadata

        """
        metadata = self.dataset.metadata.to_dict()
        metadata_prepared = self._ndarrays_to_list_recursively(metadata)
        return metadata_prepared

    def _ndarrays_to_list_recursively(self, dictionary):
        """Transforms instances of :class:`numpy.array` from a given dict.

        Numpy arrays are not easily convertible into a yaml-file and are
        therefore transformed to lists.
        This is a cascaded method and thus also works on nested dicts.

        Parameters
        ----------
        dictionary: :class:`dict`
            Dictionary to relieve of arrays.

        Returns
        -------
        dictionary: :class:`dict`
            Dictionary relieved of arrays.

        """
        for key, value in dictionary.items():
            if isinstance(value, dict):
                dictionary[key] = self._ndarrays_to_list_recursively(value)
            if isinstance(value, np.ndarray):
                dictionary[key].to_list()
        return dictionary


class MetadataExporter(aspecd.io.DatasetExporter):
    """Export metadata to yaml file with default name."""

    def __init__(self):
        super().__init__()
        self.metadata_dict = collections.OrderedDict()
        self.filename = ''

    def _export(self):
        if not self.filename:
            timestamp = datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
            self.filename = 'metadata-' + timestamp + '.yaml'
        if not self.filename.endswith(('.yaml', '.yml')):
            self.filename = self.filename + '.yaml'
        self._write_metadata()

    def _write_metadata(self):
        self.metadata_dict = self.dataset.metadata.to_dict()
        self.metadata_dict = \
            self._remove_empty_items_recursively(dict_=self.metadata_dict)
        yaml_file = aspecd.utils.Yaml()
        yaml_file.dict = self.metadata_dict
        yaml_file.write_to(self.filename)

    def _remove_empty_items_recursively(self, dict_=None):
        tmp_dict = collections.OrderedDict()
        for key, value in dict_.items():
            if isinstance(value, dict):
                dict_[key] = \
                    self._remove_empty_items_recursively(value)
            # if magnettech has not measured the q-value it is -1
            if key == 'q_value' and value == -1:
                continue
            if dict_[key]:
                tmp_dict[key] = dict_[key]
        return tmp_dict



import os
import unittest

import cwepr.io
import cwepr.dataset

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestTxtImporter(unittest.TestCase):
    def test_import(self):
        source = os.path.join(ROOTPATH, 'testdata/noisy_data')
        importer = cwepr.io.TxtImporter(source=source)
        dataset = cwepr.dataset.ExperimentalDataset()
        dataset.import_from(importer)



"""Tests for datasets."""

import unittest

from cwepr import dataset
import os

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestDataset(unittest.TestCase):
    def setUp(self):
        self.dataset = dataset.ExperimentalDataset()

    def test_instantiate_class(self):
        pass


class TestImport(unittest.TestCase):
    pass


class TestDatasetFactory(unittest.TestCase):
    def setUp(self):
        pass

    def test_dataset_factory(self):
        source = os.path.join(ROOTPATH, "io/testdata/test-bes3t-1D-fieldsweep")
        factory = dataset.DatasetFactory()
        ds = factory.get_dataset(source=source)
        assert(type(ds) == dataset.ExperimentalDataset)



import os
from unittest import TestCase

import aspecd.dataset
import numpy as np

import cwepr.analysis
import cwepr.dataset
import cwepr.io.magnettech
import unittest

import scipy.signal.windows

ROOTPATH = os.path.split(os.path.abspath(__file__))[0]


class TestAnalysis(unittest.TestCase):
    def setUp(self):
        source = os.path.join(ROOTPATH, 'io/testdata/test-bes3t-1D-fieldsweep')
        importer = cwepr.io.bes3t.BES3TImporter(source=source)
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.dataset.import_from(importer)

    def test_field_correction_value(self):
        analysator = cwepr.analysis.FieldCalibration()
        analysator.parameters['standard'] = 'dpph'
        analysator = self.dataset.analyse(analysator)
        self.assertTrue(analysator.parameters['mw_frequency'])
        self.assertTrue(analysator.parameters['g_value'] == 2.0036)
        self.assertEqual(np.float64, type(analysator.result))

    def test_area_under_curve(self):
        analysator = cwepr.analysis.AreaUnderCurve()
        analysator = self.dataset.analyse(analysator)
        self.assertEqual(np.float64, type(analysator.result))

    def test_linewidth_peak_to_peak(self):
        analysator = cwepr.analysis.LinewidthPeakToPeak()
        analysator = self.dataset.analyse(analysator)
        self.assertEqual(np.float64, type(analysator.result))

    def test_linewidth_fwhm(self):
        analysator = cwepr.analysis.LinewidthFWHM()
        analysator = self.dataset.analyse(analysator)
        self.assertEqual(np.float64, type(analysator.result))

    def test_snr(self):
        analysator = cwepr.analysis.SignalToNoiseRatio()
        analysator = self.dataset.analyse(analysator)
        self.assertEqual(np.float64, type(analysator.result))


class TestFieldCalibration(unittest.TestCase):
    def setUp(self):
        self.analysis = cwepr.analysis.FieldCalibration()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.data = np.sin(np.linspace(0, 2 * np.pi, num=500))
        self.mw_frequency = 9.68
        self.standard = 'LiLiF'
        self.center_field = 345.410996  # For LiLiF with given MW freq

    def test_instantiate_class(self):
        pass

    def test_has_appropriate_description(self):
        self.assertIn('magnetic field offset',
                      self.analysis.description.lower())

    def test_perform_without_frequency_raises(self):
        with self.assertRaisesRegex(ValueError, "microwave frequency"):
            self.dataset.analyse(self.analysis)

    def test_perform_without_standard_and_g_value_raises(self):
        self.analysis.parameters["mw_frequency"] = 9.68
        with self.assertRaisesRegex(ValueError, "standard or g value"):
            self.dataset.analyse(self.analysis)

    def test_perform_sets_parameters(self):
        self.dataset.data.data = self.data
        self.dataset.metadata.bridge.mw_frequency.value = self.mw_frequency
        self.analysis.parameters["standard"] = self.standard
        analysis = self.dataset.analyse(self.analysis)
        self.assertEqual(self.mw_frequency,
                         analysis.parameters["mw_frequency"])
        self.assertEqual(2.002293, analysis.parameters["g_value"])

    def test_perform_returns_correct_value(self):
        self.dataset.data.data = self.data
        self.dataset.data.axes[0].values = np.linspace(
            self.center_field - 1, self.center_field + 1, len(self.data)
        )
        self.dataset.metadata.bridge.mw_frequency.value = self.mw_frequency
        self.analysis.parameters["standard"] = self.standard
        analysis = self.dataset.analyse(self.analysis)
        self.assertAlmostEqual(0, analysis.result, 4)

    def test_result_has_correct_sign(self):
        self.dataset.data.data = self.data
        self.dataset.data.axes[0].values = np.linspace(
            self.center_field - 1, self.center_field + 2, len(self.data)
        )
        self.dataset.metadata.bridge.mw_frequency.value = self.mw_frequency
        self.analysis.parameters["standard"] = self.standard
        analysis = self.dataset.analyse(self.analysis)
        self.assertLess(analysis.result, 0)


class TestAmplitude(unittest.TestCase):
    def setUp(self):
        self.analysator = cwepr.analysis.Amplitude()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        self.data = np.sin(np.linspace(0, 2 * np.pi, num=500))

    def test_instantiate_class(self):
        cwepr.analysis.Amplitude()

    def test_has_appropriate_description(self):
        self.assertIn('amplitude', self.analysator.description.lower())

    def test_get_amplitude_1d_dataset(self):
        self.dataset.data.data = self.data
        analysis = self.dataset.analyse(self.analysator)
        self.assertAlmostEqual(2, analysis.result, 4)
        self.assertEqual(np.float64, type(analysis.result))

    def test_get_amplitude_2d_dataset(self):
        self.dataset.data.data = np.transpose(np.tile(self.data, (4, 1)))
        analysis = self.dataset.analyse(self.analysator)
        np.testing.assert_almost_equal(analysis.result, 2, decimal=4)
        self.assertEqual(np.ndarray, type(analysis.result))


class TestAmplitudeVsPower(unittest.TestCase):
    def setUp(self):
        self.analysator = cwepr.analysis.AmplitudeVsPower()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        data = np.sin(np.linspace(0, 2 * np.pi, num=500))
        self.dataset.data.data = np.transpose(np.tile(data, (4, 1)))
        self.dataset.data.axes[1].values = np.array([10, 5, 2.5, 1.25, 0.6125,
                                                     0.305])
        self.dataset.data.axes[1].unit = 'mW'

    def test_instantiate_class(self):
        cwepr.analysis.AmplitudeVsPower()

    def test_has_description(self):
        self.assertNotIn('abstract', self.analysator.description.lower())

    def test_calculate_dataset(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertEqual(2, len(analysis.result.data.axes))
        self.assertEqual(len(np.sqrt(self.dataset.data.axes[1].values)),
                         len(analysis.result.data.axes[0].values))
        self.assertEqual('sqrt(mW)', analysis.result.data.axes[0].unit)

    def test_returns_ascending_x_axis(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertGreater(analysis.result.data.axes[0].values[1] -
                           analysis.result.data.axes[0].values[0], 0)


class TestFitOnData(unittest.TestCase):
    def setUp(self):
        self.analysator = cwepr.analysis.FitOnData()
        self.dataset = cwepr.dataset.ExperimentalDataset()

    def test_instantiate_class(self):
        cwepr.analysis.FitOnData()

    def test_has_description(self):
        self.assertNotIn('abstract', self.analysator.description.lower())

    def test_fit_returns_coefficients(self):
        self.dataset.data.data = np.linspace(1, 21)
        self.dataset.data.axes[0].values = np.linspace(1, 11)
        analysis = self.dataset.analyse(self.analysator)
        self.assertTrue(all(analysis.result))

    def test_fit_only_takes_first_points(self):
        self.dataset.data.data = np.concatenate([np.linspace(1, 5, num=5),
                                                 np.linspace(5.1, 6.1, num=5)])
        self.dataset.data.axes[0].values = np.linspace(1, 10, num=10)
        analysis = self.dataset.analyse(self.analysator)
        self.assertAlmostEqual(1, analysis.result[0])

    def test_fit_does_second_order(self):
        self.dataset.data.axes[0].values = np.linspace(1, 10, num=10)
        self.dataset.data.data = 4 * self.dataset.data.axes[0].values ** 2
        self.analysator.parameters['order'] = 2
        analysis = self.dataset.analyse(self.analysator)
        self.assertAlmostEqual(4, analysis.result[0])

    def test_fixed_offset_with_offset_zero(self):
        self.dataset.data.data = np.asarray([0.45, 1, 2, 3, 4, 5])
        self.dataset.data.axes[0].values = np.asarray([0.5, 1, 2, 3, 4, 5])
        self.analysator.parameters['fixed_intercept'] = True
        res = self.dataset.analyse(self.analysator)
        self.assertAlmostEqual(res.result[1], 1, 2)
        self.assertIsInstance(res.result, list)
        self.assertEqual(res.result[0], 0)

    def test_fixed_offset_with_offset_non_zero(self):
        self.dataset.data.data = np.asarray([1.4, 2, 3, 4, 5, 6])
        self.dataset.data.axes[0].values = np.asarray([0.5, 1, 2, 3, 4, 5])
        self.analysator.parameters['fixed_intercept'] = True
        self.analysator.parameters['offset'] = 1
        res = self.dataset.analyse(self.analysator)
        self.assertAlmostEqual(res.result[1], 1, 2)
        self.assertEqual(res.result[0], 1)

    def test_fixed_offset_with_offset_zero_return_pol_coeffs(self):
        self.dataset.data.data = np.asarray([0.45, 1, 2, 3, 4, 5])
        self.dataset.data.axes[0].values = np.asarray([0.5, 1, 2, 3, 4, 5])
        self.analysator.parameters['fixed_intercept'] = True
        self.analysator.parameters['polynomial_coefficients'] = True
        res = self.dataset.analyse(self.analysator)
        self.assertIsInstance(res.result, list)

    def test_fit_returns_calculated_dataset_of_linear_slope(self):
        self.dataset.data.data = np.asarray([0.45, 1, 2, 3, 4, 5])
        self.dataset.data.axes[0].values = np.asarray([0.5, 1, 2, 3, 4, 5])
        self.analysator.parameters['fixed_intercept'] = True
        self.analysator.parameters['return_type'] = 'dataset'
        analysis = self.dataset.analyse(self.analysator)
        self.assertEqual(aspecd.dataset.CalculatedDataset,
                         type(analysis.result))
        self.assertEqual(len(self.dataset.data.axes[0].values),
                         analysis.result.data.data.shape[0])

    def test_fit_returns_calculated_dataset_higher_order(self):
        self.dataset.data.data = np.concatenate([np.linspace(1, 5, num=5),
                                                 np.linspace(5.1, 6.1, num=5)])
        self.dataset.data.axes[0].values = np.linspace(1, 10, num=10)
        self.analysator.parameters['return_type'] = 'dataset'
        analysis = self.dataset.analyse(self.analysator)
        self.assertEqual(aspecd.dataset.CalculatedDataset,
                         type(analysis.result))
        self.assertEqual(len(self.dataset.data.axes[0].values),
                         analysis.result.data.data.shape[0])


class TestPtpVsModAmp(TestCase):

    def setUp(self):
        self.analysator = cwepr.analysis.PtpVsModAmp()
        self.dataset = cwepr.dataset.ExperimentalDataset()
        data = np.array([])
        for sigma in np.linspace(5, 110):
            data = np.append(data, np.gradient(scipy.signal.windows.gaussian(
                1000, sigma)))
        data = data.reshape(50, 1000).T
        self.dataset.data.data = data
        self.dataset.data.axes[1].values = np.linspace(0, 5, num=50)

    def test_instantiate_class(self):
        cwepr.analysis.PtpVsModAmp()

    def test_has_description(self):
        self.assertNotIn('abstract', self.analysator.description.lower())

    def test_perform_task(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertEqual(aspecd.dataset.CalculatedDataset,
                         type(analysis.result))

    def test_result_has_no_values_in_last_axis(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertFalse(analysis.result.data.axes[1].values.size > 0)

    def test_axis_zero_has_the_correct_length(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertTrue(len(analysis.result.data.axes[0].values) == 50)

    def test_axis_zero_has_correct_values(self):
        analysis = self.dataset.analyse(self.analysator)
        self.assertTrue(analysis.result.data.axes[0].values[-1] == 5)



"""
Importer for the Bruker EMX and ESP format.

The Bruker EMX and ESP formats are used by older Bruker EPR spectrometers,
namely old EMX spectrometers running WinEPR and the ESP line of spectrometers.

A bit of a problem with these two formats is that they are quite similar,
but not the same. Namely the format of the file containing the data in
binary representation is completely different. The way to tell those
two formats apart is to import the ``.par`` file and look, if it contains
``DOS Format`` in the first line.

"""
import glob
import os
import re
from collections import OrderedDict
from datetime import datetime, timedelta

import numpy as np

import aspecd.io
import aspecd.infofile
import aspecd.annotation
import aspecd.metadata
import aspecd.utils


class ESPWinEPRImporter(aspecd.io.DatasetImporter):
    """Importer for the Bruker ESP and EMX formats.

    The Bruker EMX and ESP formats consist of two files, a data file with
    extension "spc" and a parameter file with extension "par". The
    particular problem with these two file formats is the different format
    of the binary data used. From the official specifications, there is no
    way to discriminate between ESP format (using Motorola format,
    translating to four-byte integer big endian), while the (old) EMX format
    (newer EMX machines probably use the BES3T format,
    see :class:`cwepr.io.BES3TImporter` for details) uses standard IEEE
    binary, translating to eight-byte float little endian.

    Furthermore, the parameter file usually contains only those values that
    deviate from the standard values given in the specification. Both are
    getting imported and the standard values overwritten by the differing
    values specified in the parameter file.


    .. versionadded:: 0.2

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self.load_infofile = True
        # private properties
        self._infofile = aspecd.infofile.Infofile()
        self._par_dict = {}
        self._mapper_filename = 'par_keys.yaml'
        self._metadata_dict = OrderedDict()
        self._file_encoding = ''

    def _import(self):
        self._clean_filenames()
        self._set_defaults()
        self._read_parameter_file()
        self._import_data()

        if self.load_infofile and self._infofile_exists():
            self._load_infofile()
            self._map_infofile()

        self._map_par_file()
        self._set_metadata()
        self._get_number_of_points()
        self._ensure_common_units()
        self._fill_axes()

    def _clean_filenames(self):
        # Dirty fix: Cut file extension
        if self.source.endswith((".par", ".spc")):
            self.source = self.source[:-4]

    def _set_defaults(self):
        default_file = aspecd.utils.Yaml()
        default_file.read_stream(aspecd.utils.get_package_data(
            'cwepr@io/par_defaults.yaml').encode())
        self._metadata_dict = default_file.dict

    def _read_parameter_file(self):
        par_filename = self.source + '.par'
        with open(par_filename, 'r') as file:
            lines = file.read().splitlines()

        for line in lines:
            line = line.split(maxsplit=1)
            key = line[0]
            if len(line) > 1:
                value = line[1]
            else:
                value = ''
            if re.match(r'^[+-]?[0-9.]+([eE][+-]?[0-9]*)?$', value):
                value = float(value)
            self._par_dict[key] = value

    def _import_data(self):
        complete_filename = self.source + '.spc'
        self._get_file_encoding()
        raw_data = np.fromfile(complete_filename, self._file_encoding)
        self.dataset.data.data = raw_data

    def _get_file_encoding(self):
        if ('DOS', 'Format') in self._par_dict.items():
            self._file_encoding = '<f'
        else:
            self._file_encoding = '>i4'

    def _infofile_exists(self):
        if self._get_infofile_name() and os.path.exists(
                self._get_infofile_name()[0]):
            return True
        print(f'No infofile found for dataset {os.path.split(self.source)[1]}, '
              f'import continued without infofile.')
        return False

    def _load_infofile(self):
        """Import infofile and parse it."""
        infofile_name = self._get_infofile_name()
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _get_infofile_name(self):
        return glob.glob(''.join([self.source.strip(), '.info']))

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        mapper.recipe_filename = 'cwepr@metadata_mapper_cwepr.yaml'
        mapper.map()
        infofile_dict = aspecd.utils.convert_keys_to_variable_names(
            mapper.metadata)
        aspecd.utils.copy_keys_between_dicts(infofile_dict,
                                             self._metadata_dict)
        aspecd.utils.copy_values_between_dicts(infofile_dict,
                                               self._metadata_dict)

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()

    def _map_par_file(self):
        yaml_file = aspecd.utils.Yaml()
        yaml_file.read_stream(aspecd.utils.get_package_data(
            'cwepr@io/' + self._mapper_filename).encode())
        metadata_dict = {}
        metadata_dict = self._traverse(yaml_file.dict, metadata_dict)
        #metadata_dict = self._check_if_temperature_empty(metadata_dict)
        aspecd.utils.copy_keys_between_dicts(metadata_dict, self._metadata_dict)
        aspecd.utils.copy_values_between_dicts(metadata_dict,
                                               self._metadata_dict)
        self._extract_datetime()

    # TODO: Implement handling of "RT" in temperature value
    @staticmethod
    def _check_if_temperature_empty(metadata_dict):
        print(metadata_dict['temperature_control'])
        if 'value'not in metadata_dict['temperature_control'][
            'temperature'].keys() or metadata_dict['temperature_control'][
                                     'temperature']['value'] == 0:
            metadata_dict.pop('temperature_control')
        return metadata_dict

    def _extract_datetime(self):
        start_date = self._try_parsing_date()
        if 'measurement' not in self._metadata_dict.keys():
            self._metadata_dict['measurement'] = {}
        self._metadata_dict['measurement']['start'] = str(start_date)
        if 'end' not in self._metadata_dict['measurement'].keys():
            self._metadata_dict['measurement']['end'] = \
                str(start_date + timedelta(minutes=1))

    def _try_parsing_date(self):
        date = self._par_dict['JDA'] + ' ' + self._par_dict['JTM']
        for fmt in ('%d-%b-%Y %H:%M:%S', '%d.%b.%Y %H:%M', '%m/%d/%Y %H:%M'):
            try:
                return datetime.strptime(date, fmt)
            except ValueError:
                pass
        raise ValueError('no valid date format found')

    def _set_metadata(self):
        self.dataset.metadata.from_dict(self._metadata_dict)

    def _traverse(self, dict_, metadata_dict):
        for key, value in dict_.items():
            if isinstance(value, dict):
                metadata_dict[key] = {}
                self._traverse(value, metadata_dict[key])
            elif value in self._par_dict.keys():
                metadata_dict[key] = self._par_dict[value]
            elif key == 'specified_unit':
                metadata_dict['unit'] = value
        return metadata_dict

    def _ensure_common_units(self):
        """Transform selected values and units to common units.

        Because of information are doubled from the infofile and the
        par-file, some units are wrong and are corrected manually here.
        """
        # microwave frequency
        if self.dataset.metadata.bridge.mw_frequency.value > 50:
            self.dataset.metadata.bridge.mw_frequency.value /= 1e9
        self.dataset.metadata.bridge.mw_frequency.unit = 'GHz'
        # microwave power
        if self.dataset.metadata.bridge.power.value < 0.001:
            self.dataset.metadata.bridge.power.value *= 1e3
        self.dataset.metadata.bridge.power.unit = 'mW'
        # magnetic field objects
        objects_ = ('start', 'stop', 'sweep_width')
        for object_ in objects_:
            magnetic_field_object = getattr(
                self.dataset.metadata.magnetic_field, object_)
            if object_ in ('start', 'stop'):
                if magnetic_field_object.value > 1500:
                    magnetic_field_object.unit = 'G'
                else:
                    magnetic_field_object.unit = 'mT'
            if magnetic_field_object.unit == 'G':
                magnetic_field_object.value /= 10
                magnetic_field_object.unit = 'mT'
            setattr(
                self.dataset.metadata.magnetic_field, object_,
                magnetic_field_object)
        if not self.dataset.metadata.temperature_control.temperature.unit:
            self.dataset.metadata.temperature_control.temperature.unit = 'K'

    def _fill_axes(self):
        self._get_magnetic_field_axis()
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[0].unit = \
            self.dataset.metadata.magnetic_field.start.unit
        self.dataset.data.axes[-1].quantity = 'intensity'

    def _get_magnetic_field_axis(self):
        # Abbreviations:
        start = self.dataset.metadata.magnetic_field.start.value
        points = int(self.dataset.metadata.magnetic_field.points)
        sweep_width = self.dataset.metadata.magnetic_field.sweep_width.value
        # in WinEPR, Bruker takes the number of points correctly (in contrast
        # to other formats...)
        stop = start + sweep_width
        # Set axis
        magnetic_field_axis = np.linspace(start, stop, points)
        assert len(magnetic_field_axis) == points, \
            'Length of magnetic field and number of points differ'
        assert len(magnetic_field_axis) == self.dataset.data.data.shape[0], \
            'Length of magnetic field and size of data differ'
        # set more values in dataset
        self.dataset.metadata.magnetic_field.stop.value = stop
        self.dataset.data.axes[0].values = magnetic_field_axis

    def _get_number_of_points(self):
        self.dataset.metadata.magnetic_field.points = len(
            self.dataset.data.data)
